from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas_3gs.datastructures import Mesh3gs

from compas.geometry import add_vectors


__all__ = ['EGI']


class EGI(Mesh3gs):
    """An EGI, represented as a mesh object.

    An EGI is a topological (mesh) dual of a cell of a volmesh, with all of its elements represented on a unit sphere.

    """

    def __init__(self):
        super(EGI, self).__init__()

        a = {'origin': (0, 0, 0)}
        va = {'x_fix': False,
              'y_fix': False,
              'z_fix': False,
              'type': None,
              'normal': None,
              'target_area': None}
        ea = {}
        fa = {}

        self.attributes.update(a)
        self.default_vertex_attributes.update(va)
        self.default_edge_attributes.update(ea)
        self.default_face_attributes.update(fa)

    # --------------------------------------------------------------------------
    #   constructors
    # --------------------------------------------------------------------------

    @classmethod
    def from_volmesh_cell(cls, ckey, volmesh):
        egi = cls()
        origin = volmesh.cell_centroid(ckey)
        egi.attributes['name'] = ckey
        egi.attributes['origin'] = origin

        halffaces = volmesh.cell_halffaces(ckey)
        vertices = volmesh.cell_vertices(ckey)

        for hfkey in halffaces:
            normal = volmesh.halfface_oriented_normal(hfkey)
            x, y, z = add_vectors(origin, normal)
            egi.add_vertex(key=hfkey, x=x, y=y, z=z)

        for vkey in vertices:
            face = volmesh.cell_vertex_halffaces(ckey, vkey)
            egi.add_face(face, fkey=vkey)

        return egi


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from .forcevolmeshobject import *  # noqa: F401 F403
from .formnetworkobject import *  # noqa: F401 F403
from .volmeshobject import *  # noqa: F401 F403
from .networkobject import *  # noqa: F401 F403

from compas_3gs.diagrams import ForceVolMesh
from compas_3gs.diagrams import FormNetwork

from .forcevolmeshobject import ForceVolMeshObject
from .formnetworkobject import FormNetworkObject
from .volmeshobject import VolMeshObject
from .networkobject import NetworkObject

VolMeshObject.register(ForceVolMesh, ForceVolMeshObject)
NetworkObject.register(FormNetwork, FormNetworkObject)

__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas

from compas.geometry import add_vectors
from compas.geometry import scale_vector
from compas.geometry import normalize_vector
from compas.geometry import intersection_line_plane

from compas.utilities import i_to_rgb

from compas_rhino.objects import mesh_select_face

from compas_3gs.operations import cell_split_indet_face_vertices
from compas_3gs.operations import cell_relocate_face

from compas_3gs.utilities import polygon_normal_oriented
from compas_3gs.utilities import polygon_area_oriented
from compas_3gs.utilities import datastructure_centroid

from compas_3gs.rhino import get_target_point

try:
    import Rhino
    import rhinoscriptsyntax as rs

    from Rhino.Geometry import Point3d
    from Rhino.Geometry import Vector3d
    from Rhino.Geometry import Line

    from System.Drawing.Color import FromArgb

    black = FromArgb(0, 0, 0)
    gray = FromArgb(200, 200, 200)
    white = FromArgb(255, 255, 255)

except ImportError:
    compas.raise_if_ironpython()


__all__ = ['rhino_cell_face_pull']


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   cell transformations
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def rhino_cell_face_pull(cell):

    # --------------------------------------------------------------------------
    #  1. pick face
    # --------------------------------------------------------------------------
    cell.draw()
    face = mesh_select_face(cell)
    cell_split_indet_face_vertices(cell, face)
    cell.clear()

    # --------------------------------------------------------------------------
    #  2. face data
    # --------------------------------------------------------------------------
    f_normal = cell.face_normal(face)
    f_center = cell.face_center(face)
    f_area = cell.face_area(face)
    f_vkeys = cell.face_vertices(face)

    # --------------------------------------------------------------------------
    #  3. get neighbor edges and vertex coordinates
    # --------------------------------------------------------------------------
    edges = {}
    for u in f_vkeys:
        u_nbrs = cell.vertex_neighbors(u)
        for v in u_nbrs:
            if v not in f_vkeys:
                edges[u] = v

    xyz_dict = {}
    for vkey in cell.vertex:
        if vkey not in f_vkeys:
            xyz_dict[vkey] = cell.vertex_coordinates(vkey)

    # --------------------------------------------------------------------------
    #  4. dynamic draw
    # --------------------------------------------------------------------------
    rs.EnableRedraw(True)

    def OnDynamicDraw(sender, e):

        cp = e.CurrentPoint
        plane = (cp, f_normal)
        new_pt_list = []

        for u in f_vkeys:
            v = edges[u]
            u_xyz = cell.vertex_coordinates(u)
            v_xyz = cell.vertex_coordinates(v)
            line = (u_xyz, v_xyz)
            it = intersection_line_plane(line, plane)
            xyz_dict[u] = it
            new_pt_list.append(it)

            e.Display.DrawDottedLine(
                Point3d(*u_xyz),
                Point3d(*it),
                black)

        for vkey in cell.vertex:
            xyz = cell.vertex_coordinates(vkey)
            e.Display.DrawPoint(Point3d(*xyz), 0, 5, black)

        # old normal and area --------------------------------------------------
        e.Display.DrawDot(Point3d(*f_center),
                          str(round(f_area, 3)),
                          gray, white)

        # draw original face ---------------------------------------------------
        for i in range(-1, len(f_vkeys) - 1):
            vkey1 = f_vkeys[i]
            vkey2 = f_vkeys[i + 1]
            sp = Point3d(*cell.vertex_coordinates(vkey1))
            np = Point3d(*cell.vertex_coordinates(vkey2))

            e.Display.DrawDottedLine(sp, np, black)

        # get current face info ------------------------------------------------
        areas = {}
        normals = {}
        for fkey in cell.faces():
            face_coordinates = [xyz_dict[vkey] for vkey in cell.face_vertices(fkey)]
            areas[fkey] = polygon_area_oriented(face_coordinates)
            normals[fkey] = polygon_normal_oriented(face_coordinates)

        # draw new face areas / vectors ----------------------------------------
        for fkey in cell.faces():
            area = areas[fkey]
            normal = normals[fkey]
            value = area / max(areas.values())
            color = i_to_rgb(value)
            color = FromArgb(*color)

            # draw vectors -----------------------------------------------------
            scale = 0.25
            center = datastructure_centroid(cell)
            sp = Point3d(*center)
            vector = scale_vector(normal, area * scale)
            ep = Point3d(*add_vectors(center, vector))

            e.Display.DrawArrow(Line(sp, ep), color, 20, 0)

            # draw face --------------------------------------------------------
            face_coordinates = [xyz_dict[vkey] for vkey in cell.face_vertices(fkey)]
            face_coordinates.append(face_coordinates[0])
            polygon_xyz = [Point3d(*xyz) for xyz in face_coordinates]

            e.Display.DrawPolyline(polygon_xyz, black, 2)

            if fkey == face:
                e.Display.DrawPolyline(polygon_xyz, black, 4)
                e.Display.DrawPolygon(polygon_xyz, color, filled=True)

            # display force magnitudes -----------------------------------------
            vector = add_vectors(vector,
                                 scale_vector(normalize_vector(normal), 0.75))
            xyz = add_vectors(center, vector)
            if fkey == face:
                color = black

            e.Display.DrawDot(Point3d(*xyz), str(round(area, 2)), color, white)

    # --------------------------------------------------------------------------
    #  5. input point
    # --------------------------------------------------------------------------
    ip = Point3d(*f_center)
    line = Rhino.Geometry.Line(ip, ip + Vector3d(*f_normal))
    gp = get_target_point(line, OnDynamicDraw)

    # --------------------------------------------------------------------------
    #  6. update cell coordinates
    # --------------------------------------------------------------------------
    cell_relocate_face(cell, face, gp, f_normal)

    cell.draw()


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import json

import scriptcontext as sc

import compas_rhino
from compas.utilities import DataEncoder


__commandname__ = "TGS__session_save"


def RunCommand(is_interactive):

    if '3GS' not in sc.sticky:
        compas_rhino.display_message('3GS has not been initialised yet.')
        return

    system = sc.sticky['3GS']['system']
    scene = sc.sticky['3GS']['scene']

    filepath = compas_rhino.rs.SaveFileName('save', filter=system['session.extension'], folder=system['session.dirname'])

    if not filepath:
        return

    if filepath.split('.')[-1] != system['session.extension']:
        filepath = "%s.%s" % (filepath, system['session.extension'])

    dirname, basename = os.path.split(filepath)
    filename, _ = os.path.splitext(basename)

    system['session.dirname'] = dirname
    system['session.filename'] = filename

    filepath = os.path.join(dirname, filename + '.' + system['session.extension'])

    session = {
        "data": {"form": None, "force": None},
        "scene": {"form": None, "force": None}
    }

    objects = scene.find_by_name('form')
    if objects:
        form = objects[0]
        if form:
            session['data']['form'] = form.diagram.to_data()
            session['scene']['form'] = {
                'settings': form.settings,
                'anchor': form.anchor,
                'location': form.location,
                'scale': form.scale
            }

    objects = scene.find_by_name('force')
    if objects:
        force = objects[0]
        if force:
            session['data']['force'] = force.diagram.to_data()
            session['scene']['force'] = {
                'settings': force.settings,
                'anchor': force.anchor,
                'location': force.location,
                'scale': force.scale
            }

    with open(filepath, 'w+') as f:
        json.dump(session, f, cls=DataEncoder, indent=4)


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    RunCommand(True)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from compas.geometry import convex_hull
from compas.geometry import add_vectors
from compas.geometry import scale_vector
from compas.geometry import subtract_vectors

from compas.utilities import geometric_key


__all__ = ['volmesh_ud',
           'cellnetwork_ud']


def volmesh_ud(volmesh,
               network,
               scale=0.5):
    """Computes temporary vertex coordinates for every halfface of a volmesh, for the visualisation of the unified diagram.

    Parameters
    ----------
    volmesh : VolMesh
        A volmesh object representing a polyhedral force diagram.
    network : Network
        A network object representing a polyhedral form diagram.
    scale : float
        Unified diagram scale factor,

    Returns
    -------
    halffaces : dictionary
        A dictionary of dictionaries: hfkey-{vkey: (x, y, z)}.
    prisms : dictionary
        A dictinoary of dictinoaries: uv - [face coordinates]

    Raises
    ------
    Exception
        If scale is 0, which means the unified diagram is equivalent to the polyhedral force diagram.
    Exception
        If scale is 0, which means the unified diagram is equivalent to the polyhedral form diagram.

    Notes
    -----
    - The prisms are implemented as convex hull of two halffaces for simplicity and to resolve any small geometric errors.
    - Unified diagram with a scale of 0 is equivalent to the polyhedral force diagram, while a scale of 1 is equivalent to the polyhedral form diagram.

    """

    # --------------------------------------------------------------------------
    #   0. evaluate unified diagram scale
    # --------------------------------------------------------------------------
    if scale == 0:
        raise Exception("A unified diagram with a scale of 0 is equivalent to the polyhedral force diagram.")

    if scale == 1:
        raise Exception("A unified diagram with a scale of 1 is equivalent to the polyhedral form diagram.")

    assert 0 < scale and scale < 1, "Scale needs to be between 0 and 1."

    # --------------------------------------------------------------------------
    #   1. current positions of diagrams
    # --------------------------------------------------------------------------
    volmesh_center = volmesh.centroid()
    network_center = network.datastructure_centroid()
    translation = subtract_vectors(volmesh_center, network_center)

    # --------------------------------------------------------------------------
    #   2. get base points
    # --------------------------------------------------------------------------
    base_xyz = {}

    for vkey in network.nodes():
        init_xyz = network.node_coordinates(vkey)
        base_xyz[vkey] = add_vectors(init_xyz, translation)

    # --------------------------------------------------------------------------
    #   3. compute scaled halffaces
    # --------------------------------------------------------------------------
    # halffaces = {}

    # for ckey in volmesh.cells():
    #     cell_hfs = volmesh.cell_faces(ckey)
    #     for hfkey in cell_hfs:
    #         hf_vertices = {}
    #         for vkey in volmesh.halfface_vertices(hfkey):
    #             xyz = volmesh.vertex_coordinates(vkey)
    #             arm = scale_vector(subtract_vectors(xyz, base_xyz[ckey]), scale)
    #             hf_vertices[vkey] = add_vectors(base_xyz[ckey], arm)
    #         halffaces[hfkey] = hf_vertices

    scaled_halffaces = {}
    cells = {}

    for cell in volmesh.cells():
        gkey_xyz = {}
        faces = []

        for face in volmesh.cell_faces(cell):
            new_face = []
            scaled_face_xyz = []
            for vertex in volmesh.face_vertices(face):
                xyz = volmesh.vertex_coordinates(vertex)
                arm = scale_vector(subtract_vectors(xyz, base_xyz[cell]), scale)
                scaled_xyz = add_vectors(base_xyz[cell], arm)
                gkey = geometric_key(scaled_xyz)
                gkey_xyz[gkey] = scaled_xyz
                new_face.append(gkey)
                scaled_face_xyz.append(scaled_xyz)
            scaled_halffaces[face] = scaled_face_xyz
            faces.append(new_face)

        gkey_index = dict((gkey, index) for index, gkey in enumerate(gkey_xyz))
        vertices = [list(xyz) for gkey, xyz in gkey_xyz.items()]
        scaled_faces = [[gkey_index[gkey] for gkey in face] for face in faces]

        cells[cell] = {'vertices': vertices, 'faces': scaled_faces}

    # cells = {}
    # for cell in volmesh.cells():
    #     vertices = volmesh.cell_vertices(cell)
    #     faces = volmesh.cell_faces(cell)
    #     vertex_index = dict((vertex, index) for index, vertex in enumerate(vertices))
    #     scaled_vertices = []
    #     for vertex in vertices:
    #         xyz = volmesh.vertex_coordinates(vertex)
    #         arm = scale_vector(subtract_vectors(xyz, base_xyz[cell]), scale)
    #         scaled_vertices.append(add_vectors(base_xyz[cell], arm))
    #     faces = [[vertex_index[vertex] for vertex in volmesh.halfface_vertices(face)] for face in faces]
    #     cells[cell] = {'vertices': scaled_vertices, 'faces': faces}

    # --------------------------------------------------------------------------
    #   4. compute prism faces
    # --------------------------------------------------------------------------
    # prism_faces = {}

    prism_cells = {}

    for u, v in network.edges():
        u_hfkey, v_hfkey = volmesh.cell_pair_halffaces(u, v)
        u_pts = scaled_halffaces[u_hfkey]
        v_pts = scaled_halffaces[v_hfkey]
        pts = u_pts + v_pts
        prism = convex_hull(pts)  # face as indices of pt_list
        prism_cells[(u, v)] = {'vertices': pts, 'faces': prism}

    # for u, v in network.edges():
    #     u_hfkey, v_hfkey = volmesh.cell_pair_halffaces(u, v)
    #     u_pts = halffaces[u_hfkey].values()
    #     v_pts = halffaces[v_hfkey].values()
    #     pt_list = u_pts + v_pts
    #     prism = convex_hull(u_pts + v_pts)  # face as indices of pt_list
    #     face_list = []
    #     for face in prism:
    #         face_xyz = [pt_list[i] for i in face]  # get face xyz in order
    #         face_list.append(face_xyz)
    #     prism_faces[(u, v)] = face_list
    #     prism_cells[(u, v)] = {'vertices': pt_list, 'faces': face_list}

    # --------------------------------------------------------------------------

    # return halffaces, prism_faces
    return cells, prism_cells


def cellnetwork_ud(cellnetwork):
    raise NotImplementedError


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.datastructures import VolMesh

from compas.geometry import centroid_points
from compas.geometry import normalize_vector

__all__ = ['VolMesh3gs']


class VolMesh3gs(VolMesh):
    """Inherits and extends the compas VolMesh class, such that it is more suitable for 3D graphic statics applications.

    Primarily used for polyhedral form and force diagrams.

    """

    def __init__(self):
        super(VolMesh3gs, self).__init__()

    # --------------------------------------------------------------------------
    #   deleting
    # --------------------------------------------------------------------------

    def cell_vertex_delete(self, vkey):
        '''This removes the vertex, and everything that is attached to the vertex EXCEPT the cell itself.
        '''

        if len(self.cell) > 1:
            raise ValueError('This is a multi-cell volmesh.')
        nbr_vkeys = self.vertex_neighbours(vkey)
        nbr_ckeys = self.vertex_cells(vkey)

        # delete cell info -----------------------------------------------------
        for ckey in nbr_ckeys:
            del self.cell[ckey][vkey]
            for nbr_vkey in nbr_vkeys:
                del self.cell[ckey][nbr_vkey][vkey]

        # delete halffaces -----------------------------------------------------
        halffaces = self.vertex_halffaces(vkey)
        for hfkey in halffaces:
            del self.halfface[hfkey]

        # delete planes --------------------------------------------------------
        del self.plane[vkey]
        for u in self.plane:
            for v in self.plane[u].keys():
                if v == vkey:
                    del self.plane[u][v]
                else:
                    for w in self.plane[u][v].keys():
                        if w == vkey:
                            del self.plane[u][v][w]

        # delete edges ---------------------------------------------------------
        del self.edge[vkey]
        for u in self.edge:
            if vkey in self.edge[u]:
                del self.edge[u][vkey]

        # delete the vertex itself ---------------------------------------------
        del self.vertex[vkey]

    def delete_halfface(self, hfkey):
        vertices = self.halfface_vertices(hfkey)
        for i in range(-2, len(vertices) - 2):
            u = vertices[i]
            v = vertices[i + 1]
            w = vertices[i + 2]
            del self.plane[u][v][w]
            if self.plane[w][v][u] is None:
                del self.plane[w][v][u]
        del self.halfface[hfkey]

    def delete_cell(self, ckey):

        hfkeys = self.cell_halffaces(ckey)

        for hfkey in hfkeys:
            for halfedge in self.halfface_halfedges(hfkey):
                u, v = halfedge
                # delete edges
                if v in self.edge[u]:
                    if len(self.edge_cells(u, v)) == 1:
                        del self.edge[u][v]
                if u in self.edge[v]:
                    if len(self.edge_cells(v, u)) == 1:
                        del self.edge[v][u]

        # delete vertices
        for vkey in self.cell_vertices(ckey):
            if len(self.vertex_cells(vkey)) == 1:
                del self.vertex[vkey]

        # delete halfface and planes
        for hfkey in hfkeys:
            vertices = self.halfface_vertices(hfkey)
            for i in range(-2, len(vertices) - 2):
                u = vertices[i]
                v = vertices[i + 1]
                w = vertices[i + 2]
                # delete planes
                self.plane[u][v][w] = None
                if self.plane[w][v][u] is None:
                    del self.plane[u][v][w]
                    del self.plane[w][v][u]
            del self.halfface[hfkey]

        # delete cell
        del self.cell[ckey]

    # --------------------------------------------------------------------------
    #   vertices
    # --------------------------------------------------------------------------

    def vertex_normal(self, vertex):
        """Return the normal vector at the vertex as the weighted average of the
        normals of the neighboring faces.

        Parameters
        ----------
        key : int
            The identifier of the vertex.

        Returns
        -------
        list
            The components of the normal vector.
        """
        if not self.is_vertex_on_boundary(vertex):
            return

        halffaces = []
        for halfface in self.vertex_halffaces(vertex):
            if self.is_halfface_on_boundary(halfface):
                halffaces.append(halfface)

        vectors = [self.face_normal(halfface, False) for halfface in halffaces if halfface is not None]
        return normalize_vector(centroid_points(vectors))

    def vertex_update_xyz(self, vertex, new_xyz, constrained=True):

        if constrained:
            # X
            if self.vertex_attribute(vertex, 'x_fix') is False:
                self.vertex_attribute(vertex, 'x', new_xyz[0])
            # Y
            if self.vertex_attribute(vertex, 'y_fix') is False:
                self.vertex_attribute(vertex, 'y', new_xyz[1])
            # Z
            if self.vertex_attribute(vertex, 'z_fix') is False:
                self.vertex_attribute(vertex, 'z', new_xyz[2])
        else:
            self.vertex_attribute(vertex, 'x', new_xyz[0])
            self.vertex_attribute(vertex, 'y', new_xyz[1])
            self.vertex_attribute(vertex, 'z', new_xyz[2])

    # --------------------------------------------------------------------------
    # halfface
    # --------------------------------------------------------------------------

    def boundary_halfface_manifold_neighbors(self, halfface):
        if not self.is_halfface_on_boundary(halfface):
            return

        nbrs = []
        cell = self.halfface_cell(halfface)
        for u, v in self.halfface_halfedges(halfface):
            nbr_halfface = self._cell[cell][v][u]
            w = self.halfface_vertex_ancestor(nbr_halfface, v)
            nbr_cell = self._plane[u][v][w]
            if nbr_cell is not None:
                nbr = self._cell[nbr_cell][v][u]
                nbrs.append(nbr)
        return nbrs

        # nbrs = []
        # cell = self.halfface_cell(halfface)
        # for halfedge in self.halfface_halfedges(halfface):
        #     print('halfedge', halfedge)
        #     print('edge halffaces', self.edge_halffaces(halfedge))

        #     for face in self.edge_halffaces(halfedge):
        #         print('face -------------------------', face)
        #         if self.is_halfface_on_boundary(face):
        #             print('cell', cell)
        #             print('nbr_ckey', self.halfface_cell(face))

        #             if self.halfface_cell(face) is not cell:
        #                 print('add face')
        #                 nbrs.append(face)
        # print('nbrs', nbrs)
        # return nbrs

    def boundary_halfface_manifold_neighborhood(self, hfkey, ring=1):
        nbrs = set(self.boundary_halfface_manifold_neighbors(hfkey))
        i = 1
        while True:
            if i == ring:
                break
            temp = []
            for nbr_hfkey in nbrs:
                temp += self.boundary_halfface_manifold_neighbors(hfkey)
            nbrs.update(temp)
            i += 1
        return list(nbrs - set([hfkey]))

    # --------------------------------------------------------------------------
    # cell
    # --------------------------------------------------------------------------

    def cell_pair_halffaces(self, cell_1, cell_2):
        """Given 2 ckeys, returns the interfacing halffaces, respectively.
        Parameters
        ----------
        ckey_1 : hashable
            Identifier of the cell 1.
        ckey_2 : hashable
            Identifier of the cell 2.
        Returns
        -------
        hfkey_1
            The identifier of the halfface belonging to cell 1 .
        hfkey_2
            The identifier of the halfface belonging to cell 2.
        """
        for halfface in self.cell_faces(cell_1):
            u, v, w = self.halfface_vertices(halfface)[0:3]
            nbr = self._plane[w][v][u]

            if nbr == cell_2:
                return halfface, self.halfface_opposite_halfface(halfface)

        raise KeyError

    # --------------------------------------------------------------------------
    # drawing
    # --------------------------------------------------------------------------

    # def draw(self, **kwattr):
    #     artist = VolMeshArtist(self, layer=self.layer)
    #     artist.clear_by_name()
    #     artist.draw_faces(**kwattr)
    #     artist.draw_vertices(**kwattr)

    # def clear(self, **kwattr):
    #     artist = VolMeshArtist(self, layer=self.layer)
    #     artist.clear_by_name()
    #     artist.clear_layer()

    # def draw_edges(self, **kwattr):
    #     artist = VolMeshArtist(self, **kwattr)
    #     artist.draw_edges(**kwattr)

    # def draw_faces(self, **kwattr):
    #     artist = VolMeshArtist(self, layer=self.layer)
    #     artist.draw_faces(**kwattr)

    # def clear_faces(self, **kwattr):
    #     artist = VolMeshArtist(self, layer=self.layer)
    #     artist.clear_faces(**kwattr)

    # def draw_facelabels(self, **kwattr):
    #     artist = VolMeshArtist(self, layer=self.layer)
    #     artist.draw_facelabels(**kwattr)

    # def draw_vertices(self, **kwattr):
    #     artist = VolMeshArtist(self, layer=self.layer)
    #     artist.draw_vertices(**kwattr)

    # def draw_vertexlabels(self, **kwattr):
    #     artist = VolMeshArtist(self, layer=self.layer)
    #     artist.draw_vertexlabels(**kwattr)

    # def draw_edgelabels(self, **kwattr):
    #     artist = VolMeshArtist(self, layer=self.layer)
    #     artist.draw_edgelabels(**kwattr)


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from .dynamic_pickers import *  # noqa: F401 F403
from .helpers import *  # noqa: F401 F403
from .inspectors import *  # noqa: F401 F403
from .selectors import *  # noqa: F401 F403

__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino


__commandname__ = "TGS__undo"


def RunCommand(is_interactive):

    if '3GS' not in sc.sticky:
        compas_rhino.display_message('3GS has not been initialised yet.')
        return

    scene = sc.sticky['3GS']['scene']
    if not scene:
        return

    if not scene.undo():
        compas_rhino.display_message("Nothing left to undo.")


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    RunCommand(True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

from compas.utilities import i_to_red

import compas_rhino

from compas_3gs.algorithms import volmesh_planarise

from compas_3gs.rhino import VolmeshConduit

from compas_3gs.utilities import volmesh_face_flatness
from compas_3gs.utilities import compare_initial_current


__commandname__ = "TGS_force_planarize"


def RunCommand(is_interactive):

    sc.doc.EndUndoRecord(sc.doc.CurrentUndoRecordSerialNumber)

    if '3GS' not in sc.sticky:
        compas_rhino.display_message('3GS has not been initialised yet.')
        return

    scene = sc.sticky['3GS']['scene']
    if not scene:
        return

    # get ForceVolMeshObject from scene ----------------------------------------
    objects = scene.find_by_name('force')
    if not objects:
        compas_rhino.display_message("There is no force diagram in the scene.")
        return
    force = objects[0]

    # --------------------------------------------------------------------------

    current_setting = force.settings['show.vertices']
    if not current_setting:
        force.settings['show.vertices'] = True
        scene.update()

    # get global settings ------------------------------------------------------
    kmax = scene.settings['Solvers']['planarization.kmax']
    refresh = scene.settings['Solvers']['planarization.refreshrate']
    tol = scene.settings['Solvers']['planarization.tol']

    options = ['Iterations', 'Refreshrate', 'Fix_vertices', 'Tolerance']

    # options ------------------------------------------------------------------
    fix_vertices = []

    while True:
        option = compas_rhino. rs.GetString('Press Enter to run or ESC to exit.', strings=options)

        if option is None:
            print("Planarization aborted!")
            return

        if not option:
            break

        if option == 'Iterations':
            new_kmax = compas_rhino.rs.GetInteger('Enter number of iterations', kmax, 1, 10000)
            if new_kmax or new_kmax is not None:
                kmax = new_kmax

        elif option == 'RefreshRate':
            new_refresh = compas_rhino.rs.GetInteger('Refresh rate for dynamic visualisation', refresh, 0, 1000)
            if new_refresh or new_refresh is not None:
                refresh = new_refresh

        elif option == 'Fix_vertices':
            fix_vertices += force.select_vertices()

        elif option == 'Tolerance':
            new_tol = compas_rhino.rs.GetReal('Enter flatness tolerance', tol, 0.001, 1.0)
            if new_tol or new_tol is not None:
                tol = new_tol

    if refresh > kmax:
        refresh = 0

    scene.settings['Solvers']['planarization.kmax'] = kmax
    scene.settings['Solvers']['planarization.refreshrate'] = refresh
    scene.settings['Solvers']['planarization.tol'] = tol

    # --------------------------------------------------------------------------
    # planarization
    # --------------------------------------------------------------------------

    if refresh > 0:

        initial_flatness = volmesh_face_flatness(force.diagram)

        conduit = VolmeshConduit(force.diagram)

        def callback(forcediagram, k, args, refreshrate=refresh):
            if k % refreshrate:
                return
            current_flatness = volmesh_face_flatness(forcediagram)
            face_colordict = compare_initial_current(current_flatness,
                                                     initial_flatness,
                                                     color_scheme=i_to_red)
            conduit.face_colordict = face_colordict
            conduit.redraw()

        # planarise
        with conduit.enabled():
            volmesh_planarise(force.diagram,
                              kmax=kmax,
                              fix_vkeys=fix_vertices,
                              fix_boundary_normals=False,
                              tolerance_flat=tol,
                              callback=callback,
                              print_result_info=True)

    # check if there is form diagram -------------------------------------------
    objects = scene.find_by_name('form')
    if not objects:
        force.check_eq()
        scene.update()
        return
    form = objects[0]

    # update -------------------------------------------------------------------
    force.settings['show.vertices'] = current_setting

    form.diagram.update_angle_deviations()

    form.check_eq()
    force.check_eq()

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    RunCommand(True)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from copy import deepcopy

from compas.datastructures import Mesh

from compas.datastructures.mesh import mesh_split_face

from compas.geometry import subtract_vectors
from compas.geometry import normalize_vector
from compas.geometry import length_vector
from compas.geometry import cross_vectors

from compas_rhino.artists import MeshArtist

from compas_rhino.utilities import draw_mesh

from compas_3gs.utilities import polygon_normal_oriented
from compas_3gs.utilities import polygon_area_oriented
from compas_3gs.utilities import datastructure_centroid


__all__ = ['Mesh3gs']


class Mesh3gs(Mesh):
    """Inherits and extends the compas Mesh class, such that it is more suitable for 3D graphic statics applications.

    Primarily used for the EGI.

    """

    def __init__(self):
        super(Mesh3gs, self).__init__()

    # --------------------------------------------------------------------------
    #   inherited
    # --------------------------------------------------------------------------

    mesh_split_face = mesh_split_face
    datastructure_centroid = datastructure_centroid

    def add_edge(self, u, v, attr_dict=None, **kwattr):

        attr = deepcopy(self.default_edge_attributes)
        if not attr_dict:
            attr_dict = {}
        attr_dict.update(kwattr)
        attr.update(attr_dict)

        if u not in self.vertex:
            u = self.add_vertex(u)
        if v not in self.vertex:
            v = self.add_vertex(v)

        data = self.edge[u].get(v, {})
        data.update(attr)

        self.edge[u][v] = data
        if v not in self.halfedge[u]:
            self.halfedge[u][v] = None
        if u not in self.halfedge[v]:
            self.halfedge[v][u] = None

        return u, v

    # --------------------------------------------------------------------------
    # helpers - vertices
    # --------------------------------------------------------------------------

    def vertex_update_xyz(self, vkey, xyz, constrained=True):
        if constrained:
            # X
            if self.vertex[vkey]['x_fix'] is False:
                self.vertex[vkey]['x'] = xyz[0]
            # Y
            if self.vertex[vkey]['y_fix'] is False:
                self.vertex[vkey]['y'] = xyz[1]
            # Z
            if self.vertex[vkey]['z_fix'] is False:
                self.vertex[vkey]['z'] = xyz[2]
        else:
            self.vertex[vkey]['x'] = xyz[0]
            self.vertex[vkey]['y'] = xyz[1]
            self.vertex[vkey]['z'] = xyz[2]

    # --------------------------------------------------------------------------
    # helpers - faces
    # --------------------------------------------------------------------------

    def face_area(self, fkey):
        points = self.face_coordinates(fkey)
        area = polygon_area_oriented(points)
        return area

    def face_normal(self, fkey, unitized=True):
        points = self.face_coordinates(fkey)
        normal = polygon_normal_oriented(points, unitized)
        if length_vector(normal) == 0:
            uv = subtract_vectors(points[1], points[0])
            vw = subtract_vectors(points[2], points[1])
            normal = normalize_vector(cross_vectors(uv, vw))
        return normal

    # --------------------------------------------------------------------------
    # drawing
    # --------------------------------------------------------------------------

    def draw(self, **kwattr):
        draw_mesh(self, clear_vertices=True, clear_faces=True, clear_edges=True, **kwattr)

    def clear(self, **kwattr):
        artist = MeshArtist(self)
        artist.clear_by_name()

    def draw_vertices(self, **kwattr):
        artist = MeshArtist(self)
        artist.draw_vertices(**kwattr)

    def draw_edges(self, **kwattr):
        artist = MeshArtist(self)
        artist.draw_edges(**kwattr)

    def draw_faces(self, **kwattr):
        artist = MeshArtist(self)
        artist.draw_faces(**kwattr)

    def draw_vertexlabels(self, **kwattr):
        artist = MeshArtist(self)
        artist.draw_vertexlabels(**kwattr)

    def draw_edgelabels(self, **kwattr):
        artist = MeshArtist(self)
        artist.draw_edgelabels(**kwattr)

    def draw_facelabels(self, **kwattr):
        artist = MeshArtist(self)
        artist.draw_facelabels(**kwattr)


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas

from compas.utilities import i_to_red

import compas_rhino

from compas_3gs.algorithms import volmesh_planarise

from compas_3gs.rhino import VolmeshConduit

from compas_3gs.utilities import volmesh_face_areaness
from compas_3gs.utilities import compare_initial_current

try:
    import rhinoscriptsyntax as rs
except ImportError:
    compas.raise_if_ironpython()


__commandname__ = "TGS_force_arearize"


def RunCommand(is_interactive):

    sc.doc.EndUndoRecord(sc.doc.CurrentUndoRecordSerialNumber)

    if '3GS' not in sc.sticky:
        compas_rhino.display_message('3GS has not been initialised yet.')
        return

    scene = sc.sticky['3GS']['scene']
    if not scene:
        return

    # get ForceVolMeshObject from scene ----------------------------------------
    objects = scene.find_by_name('force')
    if not objects:
        compas_rhino.display_message("There is no force diagram in the scene.")
        return
    force = objects[0]

    # --------------------------------------------------------------------------

    current_setting = force.settings['show.vertices']
    if not current_setting:
        force.settings['show.vertices'] = True
        scene.update()

    # get global settings ------------------------------------------------------
    kmax = scene.settings['Solvers']['arearization.kmax']
    refresh = scene.settings['Solvers']['arearization.refreshrate']
    tol = scene.settings['Solvers']['arearization.tol']

    options = ['Iterations', 'Refreshrate', 'Set_target_areas', 'Fix_face_normals', 'Fix_vertices', 'Tolerance']

    # options ------------------------------------------------------------------
    fix_vertices = []
    target_areas = {}
    target_text = {}
    target_colors = {}
    target_normals = {}

    while True:

        rs.EnableRedraw(True)

        option = compas_rhino. rs.GetString('Press Enter to run or ESC to exit.', strings=options)

        if option is None:
            scene.clear_layers()
            scene.clear()
            scene.update()
            print("Arearization aborted!")
            return

        if not option:
            break

        if option == 'Iterations':
            new_kmax = compas_rhino.rs.GetInteger('Enter number of iterations', kmax, 1, 10000)
            if new_kmax or new_kmax is not None:
                kmax = new_kmax

        elif option == 'RefreshRate':
            new_refresh = compas_rhino.rs.GetInteger('Refresh rate for dynamic visualisation', refresh, 0, 1000)
            if new_refresh or new_refresh is not None:
                refresh = new_refresh

        elif option == 'Set_target_areas':

            while True:

                areas = {face: round(force.diagram.face_area(face), 2) for face in force.diagram.faces() if face not in target_areas}
                force.artist.draw_facelabels(text=areas)

                rs.EnableRedraw(True)

                faces = force.select_faces()

                if not faces or faces is None:
                    break

                target_area = rs.GetReal("Enter target area value", minimum=0.1, maximum=10000.0)

                if target_area:
                    for face in faces:

                        current_area = force.diagram.face_area(face)

                        if current_area > target_area:
                            color = (255, 0, 0)
                            text = str(round(current_area, 2)) + ' ---> ' + str(round(target_area, 2))
                        else:
                            color = (0, 0, 255)
                            text = str(round(current_area, 2)) + ' ---> ' + str(round(target_area, 2))

                        target_areas[face] = target_area
                        target_text[face] = text
                        target_colors[face] = color

                scene.clear_layers()
                scene.clear()
                scene.update()

                force.artist.draw_facelabels(text=target_text, color=target_colors)

        elif option == 'Fix_face_normals':
            target_normals = {}
            items = ("Boundary_faces", "No", "Yes"), ("Interior_Faces", "No", "Yes")
            boundary, interior = rs.GetBoolean("Fix normals of...", items, (False, False))
            if boundary:
                target_normals.update({face: force.diagram.face_normal(face) for face in force.diagram.halffaces_on_boundaries()})
            if interior:
                for face in force.diagram.faces():
                    if not force.diagram.is_halfface_on_boundary(face):
                        target_normals[face] = force.diagram.face_normal(face)

        elif option == 'Fix_vertices':
            fix_vertices += force.select_vertices()

        elif option == 'Tolerance':
            new_tol = compas_rhino.rs.GetReal('Enter areaness tolerance', tol, 0.001, 1.0)
            if new_tol or new_tol is not None:
                tol = new_tol

    if refresh > kmax:
        refresh = 0

    scene.settings['Solvers']['planarization.kmax'] = kmax
    scene.settings['Solvers']['planarization.refreshrate'] = refresh
    scene.settings['Solvers']['arearization.tol'] = tol

    # --------------------------------------------------------------------------
    # planarization
    # --------------------------------------------------------------------------

    if refresh > 0:

        initial_flatness = volmesh_face_areaness(force.diagram, target_areas)

        conduit = VolmeshConduit(force.diagram)

        def callback(forcediagram, k, args, refreshrate=refresh):
            if k % refreshrate:
                return
            current_flatness = volmesh_face_areaness(force.diagram, target_areas)
            face_colordict = compare_initial_current(current_flatness,
                                                     initial_flatness,
                                                     color_scheme=i_to_red)
            conduit.face_colordict = face_colordict
            conduit.redraw()

        # planarise
        with conduit.enabled():
            volmesh_planarise(force.diagram,
                              kmax=kmax,
                              target_areas=target_areas,
                              target_normals=target_normals,
                              fix_vkeys=fix_vertices,
                              fix_boundary_normals=False,
                              tolerance_flat=scene.settings['Solvers']['planarization.tol'],
                              tolerance_area=tol,
                              callback=callback,
                              print_result_info=True)

    # check if there is form diagram -------------------------------------------
    objects = scene.find_by_name('form')
    if not objects:
        force.check_eq()
        scene.update()
        return
    form = objects[0]

    # update -------------------------------------------------------------------
    force.settings['show.vertices'] = current_setting

    form.diagram.update_angle_deviations()

    form.check_eq()
    force.check_eq()

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    RunCommand(True)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from compas.geometry import distance_point_point
from compas.geometry import centroid_points
from compas.geometry import dot_vectors
from compas.geometry import project_point_plane

from compas_3gs.utilities import scale_polygon

from compas_3gs.operations import cell_collapse_short_edge

from compas_3gs.utilities import print_result


__all__ = ['cell_planarise',
           'volmesh_planarise']


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   mesh planarisation
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def cell_planarise(cell,
                   kmax=100,

                   target_centers={},
                   target_normals={},
                   target_areas={},

                   fix_vkeys=[],

                   avg_fkeys=[],

                   collapse_edge_length=0.1,

                   tolerance_flat=0.001,
                   tolerance_area=0.001,

                   callback=None,
                   callback_args=None,

                   print_result_info=False):
    """Planarise the faces of a cell.

    Planarisation of a cell is implemented as a three-step iterative procedure.
    At every iteration, each face is first individually projected to its best-fit plane (unless a target normal is given).
    Then, each face is re-sized to its target area (if given).
    Finally, the new vertex coordinates are computed by taking the centroid of the disconnected corners of the faces.

    Parameters
    ----------

    cell : Mesh
        A mesh object

    kmax : int, optional [100]
        Number of iterations.

    target_face_areas : dictionary, optional [{}]
        A dictionary of fkeys and target areas.
    target_face_normals : dictionary, optional [{}]
        A dictionary of fkeys and target face normals.
    target_face_centers : dictionary, optional [{}]
        A dictionary of fkeys and target face centers.

    fix_vkeys : list, optional [[]]
        List of vkeys to omit from arearisation.

    avg_fkeys : list, optional [[]]
        List of fkeys to average areas.

    collapse_edge_length : real, optional [0.1]
        Minimum length of edge to collapse.

    tolerance_flat: float, optional
        Convergence tolerance for face flatness.
    tolerance_area: float, optional
        Convergence tolerance for face areas.

    callback : callable, optional
        A user-defined callback function to be executed after every iteration.
        Default is ``None``.
    callback_args : tuple, optional
        Additional parameters to be passed to the callback.
        Default is ``None``.

    Raises
    ------
    Exception
        If a callback is provided, but it is not callable.

    .. seealso ::
        `compas.geometry.mesh_planarize_faces`

    """
    if callback:
        if not callable(callback):
            raise Exception('Callback is not callable.')

    # --------------------------------------------------------------------------
    #   1. initialise
    # --------------------------------------------------------------------------
    free_vkeys = list(set(cell.vertex) - set(fix_vkeys))

    # --------------------------------------------------------------------------
    #   2. loop
    # --------------------------------------------------------------------------
    for k in range(kmax):

        deviation_flat = 0
        deviation_area = 0
        deviation_perp = 0

        new_xyz = {vkey: [] for vkey in cell.vertex}

        # faces to be averaged -------------------------------------------------
        if avg_fkeys:
            avg_face_area = _avg_face_area(cell, avg_fkeys)

        for fkey in cell.faces():

            # evaluate current face --------------------------------------------
            f_vkeys = cell.face_vertices(fkey)
            f_v_xyz = cell.face_coordinates(fkey)
            f_normal = cell.face_normal(fkey)
            f_area = cell.face_area(fkey)
            f_center = cell.face_centroid(fkey)

            if fkey in target_centers:
                f_center = target_centers[fkey]

            if fkey in target_normals:
                target_normal = target_normals[fkey]

                # perpness deviation
                perpness = 1 - abs(dot_vectors(f_normal, target_normal))
                if perpness > deviation_perp:
                    deviation_perp = perpness

                f_normal = target_normal

            # projection plane -------------------------------------------------
            plane = (f_center, f_normal)

            # ------------------------------------------------------------------
            #   3. planarise
            # ------------------------------------------------------------------
            projected_face = []
            for xyz in f_v_xyz:
                projected_xyz = project_point_plane(xyz, plane)
                projected_face.append(projected_xyz)

                # planarisation deviation
                dist = distance_point_point(xyz, projected_xyz)
                if dist > deviation_flat:
                    deviation_flat = dist

            # ------------------------------------------------------------------
            #   4. arearise
            # ------------------------------------------------------------------
            if fkey in target_areas:
                target_area = target_areas[fkey]

            else:
                target_area = f_area

            if fkey in avg_fkeys:
                target_area = avg_face_area

            # scale factor -----------------------------------------------------
            if target_area != 0:
                scale = (target_area / f_area) ** 0.5

            elif target_area == 0:
                scale = 1 - f_area * 0.1
                # scale = (target_area / f_area) ** 0.5
                # scale = 0.9

            # scale ------------------------------------------------------------
            scaled_face = scale_polygon(projected_face, scale)

            # arearisation deviation
            areaness = abs(f_area - target_area)
            if areaness > deviation_area:
                deviation_area = areaness

            # collect new coordinates
            for i in range(len(f_vkeys)):
                new_xyz[f_vkeys[i]].append(scaled_face[i])

        # ----------------------------------------------------------------------
        #   5. compute new cell vertex coordinates
        # ----------------------------------------------------------------------
        for vkey in free_vkeys:
            final_xyz = centroid_points(new_xyz[vkey])
            cell.vertex_update_xyz(vkey, final_xyz)

        for u, v in cell.edges():
            cell_collapse_short_edge(cell,
                                     u,
                                     v,
                                     min_length=collapse_edge_length)

        # ----------------------------------------------------------------------
        #   7. check convergence
        # ----------------------------------------------------------------------
        if deviation_flat < tolerance_flat and deviation_area < tolerance_area:

            if print_result_info:

                name = "Cell planarisation"
                deviation = deviation_flat

                if target_areas:
                    name = "Cell arearisation"
                    deviation = deviation_area

                print_result(name, k, deviation)

            break

        # callback / conduit ---------------------------------------------------
        if callback:
            callback(cell, k, callback_args)


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh planarisation
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def volmesh_planarise(volmesh,
                      kmax=100,

                      target_centers={},
                      target_normals={},
                      target_areas={},

                      fix_vkeys=[],

                      fix_boundary_normals=False,
                      fix_all_normals=False,

                      min_area=None,
                      max_area=None,

                      tolerance_flat=0.001,
                      tolerance_area=0.001,
                      tolerance_perp=0.001,

                      callback=None,
                      callback_args=None,

                      print_result_info=False):
    """Planarises the halffaces of a volmesh.

    Planarisation of a volmesh is implemented as a three-step iterative procedure.
    At every iteration, each halfface is first individually projected to its best-fit plane (unless a target normal is given).
    Then, each halfface is re-sized to its target area (if given).
    Finally, the new vertex coordinates are computed by taking the centroid of the disconnected corners of the halffaces.

    Parameters
    ----------
    volmesh : VolMesh
        A volmesh object.
    kmax : int, optional [100]
        Number of iterations.
    target_face_areas : dictionary, optional [{}]
        A dictionary of fkeys and target areas.
    target_face_normals : dictionary, optional [{}]
        A dictionary of fkeys and target face normals.
    target_face_centers : dictionary, optional [{}]
        A dictionary of fkeys and target face centers.
    omit_vkeys : list, optional [[]]
        List of vkeys to omit from arearisation.
    fix_boundary_face_normals : boolean, optional [False]
        Whether to keep the initial normals of the bondary faces.
    fix_all_face_normals : boolean, optional [False]
        Whether to keep the initial normals of all faces.
    tolerance_flat: float, optional
        Convergence tolerance for face flatness.
    tolerance_area: float, optional
        Convergence tolerance for face areas against target areas.
    tolerance_perp: float, optional
        Convergence tolerance for face perpendicularity against target normals.
    callback : callable, optional
        A user-defined callback function to be executed after every iteration.
        Default is ``None``.
    callback_args : tuple, optional
        Additional parameters to be passed to the callback.
        Default is ``None``.
    print_result_info : bool, optional
        If True, print the result of the algorithm.

    Raises
    ------
    Exception
        If a callback is provided, but it is not callable.

    .. seealso ::
        `compas.geometry.mesh_planarize_faces`

    """
    if callback:
        if not callable(callback):
            raise Exception('Callback is not callable.')

    # --------------------------------------------------------------------------
    #   1. initialise
    # --------------------------------------------------------------------------
    free_vkeys = list(set(volmesh.vertices()) - set(fix_vkeys))
    initial_normals = _get_current_volmesh_normals(volmesh)
    boundary_fkeys = volmesh.halffaces_on_boundaries()

    # --------------------------------------------------------------------------
    #   2. loop
    # --------------------------------------------------------------------------
    for k in range(kmax):

        deviation_flat = 0
        deviation_area = 0
        deviation_perp = 0

        new_xyz = {vkey: [] for vkey in volmesh.vertices()}

        for fkey in volmesh.faces():

            fkey_pair = volmesh.halfface_opposite_halfface(fkey)

            # evaluate current face --------------------------------------------
            f_vkeys = volmesh.halfface_vertices(fkey)
            f_v_xyz = volmesh.halfface_coordinates(fkey)
            f_center = volmesh.halfface_center(fkey)
            f_normal = volmesh.halfface_normal(fkey)
            f_area = volmesh.halfface_area(fkey)

            # override with manual target values -------------------------------
            if _pair_membership(fkey, fkey_pair, target_centers):
                f_center = target_centers[fkey]

            if _pair_membership(fkey, fkey_pair, target_normals):
                target_normal = target_normals[fkey]

                # perpness deviation
                perpness = 1 - abs(dot_vectors(f_normal, target_normal))
                if perpness > deviation_perp:
                    deviation_perp = perpness

                f_normal = target_normal

            if fix_boundary_normals:
                if fkey in boundary_fkeys:
                    f_normal = initial_normals[fkey]['normal']
            if fix_all_normals:
                f_normal = initial_normals[fkey]['normal']

            # projection plane -------------------------------------------------
            plane = (f_center, f_normal)

            # ------------------------------------------------------------------
            #   3. planarise
            # ------------------------------------------------------------------
            new_face = []
            for xyz in f_v_xyz:
                projected_xyz = project_point_plane(xyz, plane)
                new_face.append(projected_xyz)

                # planarisation deviation
                flatness = distance_point_point(xyz, projected_xyz)
                if flatness > deviation_flat:
                    deviation_flat = flatness

            # ------------------------------------------------------------------
            #   4. arearise
            # ------------------------------------------------------------------
            if target_areas:
                if fkey in target_areas:
                    target_area = target_areas[fkey]
                    scale = (target_area / f_area) ** 0.5

                    # scale
                    new_face = scale_polygon(new_face, scale)

                    # arearisation deviation
                    areaness = abs(f_area - target_area)
                    if areaness > deviation_area:
                        deviation_area = areaness

            # collect new coordinates ------------------------------------------
            for i in range(len(f_vkeys)):
                new_xyz[f_vkeys[i]].append(new_face[i])

        # ----------------------------------------------------------------------
        #   5. compute new volmesh vertex coordinates
        # ----------------------------------------------------------------------
        for vkey in free_vkeys:
            final_xyz = centroid_points(new_xyz[vkey])
            volmesh.vertex_update_xyz(vkey, final_xyz)

        # ----------------------------------------------------------------------
        #   6. check convergence
        # ----------------------------------------------------------------------
        if deviation_flat < tolerance_flat and deviation_area < tolerance_area and deviation_perp < tolerance_perp:

            if print_result_info:

                name = "Volmesh planarisation"
                deviation = deviation_flat

                if target_areas:
                    name = "Volmesh arearisation"
                    deviation = deviation_area

                print_result(name, k, deviation)

            break

        # callback / conduit ---------------------------------------------------
        if callback:
            callback(volmesh, k, callback_args)


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   planarisation helpers
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def _pair_membership(item1, item2, container):
    if item1 in container or item2 in container:
        return True
    else:
        return False


def _get_current_mesh_normals(mesh):
    normal_dict = {}
    for fkey in mesh.face():
        center = mesh.face_center(fkey)
        normal = mesh.face_normal(fkey)
        normal_dict[fkey] = {'normal': normal, 'center': center}
    return normal_dict


def _get_current_volmesh_normals(volmesh):
    normal_dict = {}
    for hfkey in volmesh.halffaces():
        center = volmesh.halfface_center(hfkey)
        normal = volmesh.halfface_normal(hfkey)
        normal_dict[hfkey] = {'normal': normal, 'center': center}
    return normal_dict


def _avg_face_area(mesh, fkeys):
    total_area = sum(mesh.face_area(fkey) for fkey in fkeys)
    return total_area / len(fkeys)


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from .modify_diagrams import *  # noqa: F401 F403
from .transform_cell import *  # noqa: F401 F403
from .transform_volmesh import *  # noqa: F401 F403

__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas
from compas.utilities import i_to_rgb

from compas_3gs.rhino.control.inspectors import VolmeshCellInspector
from compas_3gs.rhino.control.selectors import CellSelector

try:
    import rhinoscriptsyntax as rs
except ImportError:
    compas.raise_if_ironpython()


__all__ = ['volmesh3gs_select_cell']


def volmesh3gs_select_cell(volmesh):

    cell_colors = {}
    ckeys = volmesh.cell.keys()
    for index, ckey in enumerate(ckeys):
        value = 0
        if len(ckeys) > 1:
            value = float(index) / (len(ckeys) - 1)
        color = i_to_rgb(value)
        cell_colors[ckey] = color

    volmesh.draw_cell_labels(color_dict=cell_colors)

    rs.EnableRedraw(True)

    # dynamic selector
    cell_inspector = VolmeshCellInspector(volmesh, color_dict=cell_colors)
    cell_inspector.enable()
    ckey = CellSelector.select_cell(volmesh)
    cell_inspector.disable()
    del cell_inspector

    volmesh.clear_cell_labels()

    return ckey


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino

from compas_3gs.rhino import SettingsForm
from compas_3gs.rhino import ForceVolMeshObject
from compas_3gs.rhino import FormNetworkObject


__commandname__ = "TGS_settings"


def RunCommand(is_interactive):

    if '3GS' not in sc.sticky:
        compas_rhino.display_message('3GS has not been initialised yet.')
        return

    scene = sc.sticky['3GS']['scene']

    SettingsForm.from_scene(scene, object_types=[ForceVolMeshObject, FormNetworkObject], global_settings=['3GS', 'Solvers'])

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    RunCommand(True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas

from compas.geometry import length_vector
from compas.geometry import cross_vectors
from compas.geometry import subtract_vectors

from compas.utilities import i_to_rgb

from compas_rhino.conduits import BaseConduit
from compas_rhino.ui import Mouse


try:
    from Rhino.Geometry import Point3d

    from System.Drawing.Color import FromArgb

except ImportError:
    compas.raise_if_ironpython()


__all__ = ['VolmeshVertexInspector',
           'VolmeshHalffaceInspector',
           'VolmeshCellInspector',
           'BiCellInspector']


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh vertex inspector
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


class VolmeshVertexInspector(BaseConduit):

    def __init__(self, volmesh, tol=0.1, **kwargs):
        super(VolmeshVertexInspector, self).__init__(**kwargs)

        dotcolor = (255, 0, 0)
        textcolor = (255, 255, 255)

        self.volmesh = volmesh
        self.tol = tol
        self.mouse = Mouse()
        self.dotcolor = FromArgb(*dotcolor)
        self.textcolor = FromArgb(*textcolor)

    def enable(self):
        self.mouse.Enabled = True
        self.Enabled = True

    def disable(self):
        self.mouse.Enabled = False
        self.Enabled = False

    def DrawForeground(self, e):
        p1 = self.mouse.p1
        p2 = self.mouse.p2
        v12 = subtract_vectors(p2, p1)
        l12 = length_vector(v12)

        for index, (key, attr) in enumerate(self.volmesh.vertices(True)):
            p0 = attr['x'], attr['y'], attr['z']
            text = str(index)
            v01 = subtract_vectors(p1, p0)
            v02 = subtract_vectors(p2, p0)
            l = length_vector(cross_vectors(v01, v02))

            if l12 == 0.0 or (l / l12) < self.tol:
                point = Point3d(*p0)
                e.Display.DrawDot(point, text, self.dotcolor, self.textcolor)

                break


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh halfface inspector
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


class VolmeshHalffaceInspector(BaseConduit):

    def __init__(self, volmesh, hfkeys=None, dependents=False, tol=1, **kwargs):
        super(VolmeshHalffaceInspector, self).__init__(**kwargs)

        dotcolor = (0, 0, 0)
        textcolor = (255, 255, 255)
        edgecolor = (255, 255, 0)
        facecolor = (255, 0, 0)

        self.volmesh = volmesh
        self.hfkeys = hfkeys
        self.dependents = dependents
        self.tol = tol
        self.mouse = Mouse()
        self.dotcolor = FromArgb(*dotcolor)
        self.textcolor = FromArgb(*textcolor)
        self.edgecolor = FromArgb(*edgecolor)
        self.facecolor = FromArgb(*facecolor)

    def enable(self):
        self.mouse.Enabled = True
        self.Enabled = True

    def disable(self):
        self.mouse.Enabled = False
        self.Enabled = False

    def DrawForeground(self, e):
        p1 = self.mouse.p1
        p2 = self.mouse.p2

        v12 = subtract_vectors(p2, p1)
        l12 = length_vector(v12)

        hfkeys = self.volmesh.faces()
        if self.hfkeys:
            hfkeys = self.hfkeys

        for hfkey in hfkeys:

            p0 = self.volmesh.halfface_center(hfkey)
            v01 = subtract_vectors(p1, p0)
            v02 = subtract_vectors(p2, p0)
            l = length_vector(cross_vectors(v01, v02))

            if l12 == 0.0 or (l / l12) < self.tol:

                face_coordinates = self.volmesh.halfface_coordinates(hfkey)
                face_coordinates.append(face_coordinates[0])
                polygon_xyz = [Point3d(*xyz) for xyz in face_coordinates]
                e.Display.DrawPolyline(polygon_xyz, self.edgecolor, 6)

                if self.dependents:

                    d_hfkeys = self.volmesh.halfface_manifold_neighborhood(hfkey, ring=50)

                    for d_hfkey in d_hfkeys:
                        face_coordinates = self.volmesh.halfface_coordinates(d_hfkey)
                        face_coordinates.append(face_coordinates[0])
                        polygon_xyz = [Point3d(*xyz) for xyz in face_coordinates]
                        e.Display.DrawPolyline(polygon_xyz, self.edgecolor, 2)

                break


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh cell inspector
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


class VolmeshCellInspector(BaseConduit):

    def __init__(self, volmesh, color_dict=None, tol=1, **kwargs):
        super(VolmeshCellInspector, self).__init__(**kwargs)

        dotcolor = (0, 0, 0)
        textcolor = (255, 255, 255)
        edgecolor = (0, 0, 0)
        facecolor = (255, 0, 0)

        self.volmesh = volmesh
        self.color_dict = color_dict
        self.tol = tol
        self.mouse = Mouse()
        self.dotcolor = FromArgb(*dotcolor)
        self.textcolor = FromArgb(*textcolor)
        self.edgecolor = FromArgb(*edgecolor)
        self.facecolor = FromArgb(*facecolor)

    def enable(self):
        self.mouse.Enabled = True
        self.Enabled = True

    def disable(self):
        self.mouse.Enabled = False
        self.Enabled = False

    def DrawForeground(self, e):
        p1 = self.mouse.p1
        p2 = self.mouse.p2
        v12 = subtract_vectors(p2, p1)
        l12 = length_vector(v12)

        # force diagram
        for ckey in self.volmesh.cell:
            p0 = self.volmesh.cell_center(ckey)
            v01 = subtract_vectors(p1, p0)
            v02 = subtract_vectors(p2, p0)
            l = length_vector(cross_vectors(v01, v02))
            color = self.edgecolor
            if self.color_dict:
                color = FromArgb(*self.color_dict[ckey])
            if l12 == 0.0 or (l / l12) < self.tol:
                for hfkey in self.volmesh.cell_halffaces(ckey):
                    vkeys = self.volmesh.halfface_vertices(hfkey)
                    face_coordinates = [self.volmesh.vertex_coordinates(vkey) for vkey in vkeys]
                    face_coordinates.append(face_coordinates[0])
                    polygon_xyz = [Point3d(*xyz) for xyz in face_coordinates]
                    e.Display.DrawPolyline(polygon_xyz, color, 3)
                break


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh cell - network vertex - inspector
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


class BiCellInspector(BaseConduit):

    def __init__(self, volmesh, network, tol=2, **kwargs):
        super(BiCellInspector, self).__init__(**kwargs)

        dotcolor = (0, 0, 0)
        textcolor = (255, 255, 255)
        edgecolor = (0, 0, 0)
        facecolor = (255, 0, 0)

        self.volmesh = volmesh
        self.network = network
        self.tol = tol
        self.mouse = Mouse()
        self.dotcolor = FromArgb(*dotcolor)
        self.textcolor = FromArgb(*textcolor)
        self.edgecolor = FromArgb(*edgecolor)
        self.facecolor = FromArgb(*facecolor)

        # self.color_dict = {}

        # for index, hfkey in self.volmesh.halffaces.keys():
        #     value = index / max(self.volmesh.halffaces.keys())
        #     color  = i_to_rgb(value)
        #     color  = System.Drawing.Color.FromArgb(*color)
        #     self.color_dict[hfkey] = color

    def enable(self):
        self.mouse.Enabled = True
        self.Enabled = True

    def disable(self):
        self.mouse.Enabled = False
        self.Enabled = False

    def DrawForeground(self, e):
        p1 = self.mouse.p1
        p2 = self.mouse.p2
        v12 = subtract_vectors(p2, p1)
        l12 = length_vector(v12)

        # force diagram

        for ckey in self.volmesh.cell:
            p0 = self.volmesh.cell_center(ckey)
            dual_p0 = self.network.vertex_coordinates(ckey)
            v01 = subtract_vectors(p1, p0)
            v02 = subtract_vectors(p2, p0)
            l = length_vector(cross_vectors(v01, v02))

            if l12 == 0.0 or (l / l12) < self.tol:

                hf_colors = {}

                nbr_vkeys = self.network.vertex_neighbours(ckey)

                for index, nbr_vkey in enumerate(nbr_vkeys):
                    value = float(index) / (len(nbr_vkeys) - 1)
                    color = i_to_rgb(value)
                    color = FromArgb(*color)
                    nbr_xyz = self.network.vertex_coordinates(nbr_vkey)
                    e.Display.DrawLine(Point3d(*dual_p0), Point3d(*nbr_xyz), color, 4)

                    hfkey = self.volmesh.cell_pair_halffaces(ckey, nbr_vkey)[0]
                    hf_colors[hfkey] = color

                for hfkey in self.volmesh.cell_halffaces(ckey):
                    vkeys = self.volmesh.halfface_vertices(hfkey)
                    face_coordinates = [self.volmesh.vertex_coordinates(vkey) for vkey in vkeys]
                    face_coordinates.append(face_coordinates[0])
                    polygon_xyz = [Point3d(*xyz) for xyz in face_coordinates]
                    e.Display.DrawDot(Point3d(*p0), str(ckey), self.dotcolor, self.textcolor)
                    e.Display.DrawPolyline(polygon_xyz, self.edgecolor, 2)

                    e.Display.DrawDot(Point3d(*dual_p0), str(ckey), self.dotcolor, self.textcolor)

                    if hfkey in hf_colors:
                        hf_color = hf_colors[hfkey]
                        e.Display.DrawPolygon(polygon_xyz, hf_color, filled=True)

                break


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import print_function

from compas.datastructures import Network

from compas.geometry import bounding_box

from compas.geometry import subtract_vectors
from compas.geometry import normalize_vector
from compas.geometry import length_vector

from compas_3gs.utilities import datastructure_centroid


__all__ = ['Network3gs']


class Network3gs(Network):
    """Inherits and extends the compas Network class, such that it is more suitable for 3D graphic statics applications.

    Primarily used for polyhedral (and possibly non-polyhedral) form diagrams.

    """

    def __init__(self):
        super(Network3gs, self).__init__()

    # --------------------------------------------------------------------------
    #   inherited functions
    # --------------------------------------------------------------------------

    datastructure_centroid = datastructure_centroid
    vertex_coordinates = Network.node_coordinates
    vertex_neighbors = Network.neighbors
    vertices = Network.nodes

    # --------------------------------------------------------------------------
    # misc
    # --------------------------------------------------------------------------

    def bounding_box(self):

        xyz = self.nodes_attributes('xyz', keys=list(self.nodes()))

        # x_sorted = sorted(xyz, key=lambda k: k[0])
        # y_sorted = sorted(xyz, key=lambda k: k[1])
        # z_sorted = sorted(xyz, key=lambda k: k[2])

        # x = abs(x_sorted[0][0] - x_sorted[-1][0])
        # y = abs(y_sorted[0][1] - y_sorted[-1][1])
        # z = abs(z_sorted[0][2] - z_sorted[-1][2])

        return bounding_box(xyz)

    # --------------------------------------------------------------------------
    # helpers - vertices
    # --------------------------------------------------------------------------

    def vertex_update_xyz(self, node, new_xyz, constrained=True):

        if constrained:
            # X
            if self.node_attribute(node, 'x_fix') is False:
                self.node_attribute(node, 'x', new_xyz[0])
            # Y
            if self.node_attribute(node, 'y_fix') is False:
                self.node_attribute(node, 'y', new_xyz[1])
            # Z
            if self.node_attribute(node, 'z_fix') is False:
                self.node_attribute(node, 'z', new_xyz[2])
        else:
            self.node_attribute(node, 'x', new_xyz[0])
            self.node_attribute(node, 'y', new_xyz[1])
            self.node_attribute(node, 'z', new_xyz[2])

    # --------------------------------------------------------------------------
    # helpers - edges
    # --------------------------------------------------------------------------

    def edge_vector(self, u, v, unitized=True):
        u_xyz = self.vertex_coordinates(u)
        v_xyz = self.vertex_coordinates(v)
        vector = subtract_vectors(v_xyz, u_xyz)
        if unitized:
            return normalize_vector(vector)
        return vector

    def edge_avg_length(self):
        sum_length = 0
        edge_count = 0
        for u, v in self.edges_iter():
            edge_vector = self.edge_vector(u, v, unitized=False)
            sum_length += length_vector(edge_vector)
            edge_count += 1
        return sum_length / edge_count

    # --------------------------------------------------------------------------
    # drawing
    # --------------------------------------------------------------------------

    # def draw(self, **kwattr):
    #     artist = NetworkArtist(self)
    #     artist.draw_edges(**kwattr)
    #     artist.draw_nodes(**kwattr)

    # def clear(self, **kwattr):
    #     artist = NetworkArtist(self)
    #     artist.clear_by_name()
    #     artist.clear_layer()

    # def draw_nodes(self, **kwattr):
    #     artist = NetworkArtist(self)
    #     artist.draw_nodes(**kwattr)

    # def draw_edges(self, **kwattr):
    #     artist = NetworkArtist(self)
    #     artist.draw_edges(**kwattr)

    # def draw_vertexlabels(self, **kwattr):
    #     artist = NetworkArtist(self)
    #     artist.draw_vertexlabels(**kwattr)

    # def draw_edgelabels(self, **kwattr):
    #     artist = NetworkArtist(self)
    #     artist.draw_edgelabels(**kwattr)


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino

from compas_3gs.diagrams import FormNetwork

from compas.geometry import Translation

from compas_3gs.algorithms import volmesh_dual_network

from compas_3gs.rhino import relocate_formdiagram


__commandname__ = "TGS_primal"


def RunCommand(is_interactive):

    if '3GS' not in sc.sticky:
        compas_rhino.display_message('3GS has not been initialised yet.')
        return

    scene = sc.sticky['3GS']['scene']

    # get ForceVolMeshObject from scene
    objects = scene.find_by_name('force')
    if not objects:
        compas_rhino.display_message("There is no force diagram in the scene.")
        return
    force = objects[0]

    # make FormNetwork
    form = volmesh_dual_network(force.diagram, cls=FormNetwork)

    # set dual/primal
    form.dual = force.diagram
    force.diagram.primal = form

    # add FormNetworkObject
    translation = relocate_formdiagram(force.diagram, form)
    form.transform(Translation.from_vector(translation))
    form.update_angle_deviations()

    scene.add_formnetwork(form, name='form', layer='3GS::FormDiagram')

    # form
    objects = scene.find_by_name('form')
    form = objects[0]

    force.check_eq()
    form.check_eq()

    # update
    scene.update()
    scene.save()

    print('Primal diagram successfully created.')


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":

    RunCommand(True)



"""
********************************************************************************
compas_3gs.datastructures
********************************************************************************

.. currentmodule:: compas_3gs.datastructures


network3gs
==========

inherits compas mesh and adds these functions...

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Network3gs

|

____


mesh3gs
=======

inherits compas mesh and adds these functions...

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Mesh3gs

|

____


volmesh3gs
==========

inherits compas mesh and adds these functions...

.. autosummary::
    :toctree: generated/
    :nosignatures:

    VolMesh3gs

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .mesh3gs import *  # noqa: F401 F403
from .network3gs import *  # noqa: F401 F403
from .volmesh3gs import *  # noqa: F401 F403

__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas_rhino.objects import VolMeshObject


__all__ = ['VolMeshObject']


class VolMeshObject(VolMeshObject):
    """A customised `VolMeshArtist` for 3GS `VolMesh`-based data structures."""

    @property
    def diagram(self):
        """The diagram associated with the object."""
        return self.volmesh

    @diagram.setter
    def diagram(self, diagram):
        self.volmesh = diagram



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.geometry import Point
from compas.geometry import Scale
from compas.geometry import Translation
from compas.geometry import Rotation

from compas.utilities import i_to_rgb

import compas_rhino

from compas_3gs.rhino.objects.networkobject import NetworkObject

from compas_3gs.utilities import get_force_colors_uv

__all__ = ['FormNetworkObject']


class FormNetworkObject(NetworkObject):
    """A form network object represents a polyhedral form diagram in the Rhino view.
    """

    SETTINGS = {
        'layer': "3GS::FormDiagram",

        '_is.valid': False,

        'show.nodes': True,
        'show.edges': True,
        'show.loads': True,
        'show.pipes': False,

        'color.invalid': (100, 255, 100),

        'color.node': (0, 0, 0),
        'color.nodes:is_fixed': (0, 0, 255),

        'color.edges': (0, 0, 0),
        'color.pipes': (0, 0, 0),

        'scale.loads': 0.100,
        'scale.pipes': 0.100,

        'tol.loads': 1e-3,
        'tol.pipes': 1e-3
    }

    def __init__(self, diagram, *args, **kwargs):
        super(FormNetworkObject, self).__init__(diagram, *args, **kwargs)
        self.settings.update(FormNetworkObject.SETTINGS)
        settings = kwargs.get('settings') or {}
        if settings:
            self.settings.update(settings)
        self._guid_loads = {}

    @property
    def node_xyz(self):
        """dict : The view coordinates of the network object."""
        origin = Point(0, 0, 0)
        if self.anchor is not None:
            xyz = self.diagram.node_attributes(self.anchor, 'xyz')
            point = Point(* xyz)
            T1 = Translation.from_vector(origin - point)
            S = Scale.from_factors([self.scale] * 3)
            R = Rotation.from_euler_angles(self.rotation)
            T2 = Translation.from_vector(self.location)
            X = T2 * R * S * T1
        else:
            S = Scale.from_factors([self.scale] * 3)
            R = Rotation.from_euler_angles(self.rotation)
            T = Translation.from_vector(self.location)
            X = T * R * S
        network = self.diagram.transformed(X)
        node_xyz = {node: network.node_attributes(node, 'xyz') for node in network.nodes()}
        return node_xyz

    @property
    def guid_loads(self):
        return self._guid_loads

    @guid_loads.setter
    def guid_loads(self, values):
        self._guid_loads = dict(values)

    @property
    def guid_pipe(self):
        return self._guid_pipe

    @guid_pipe.setter
    def guid_pipe(self, values):
        self._guid_pipe = dict(values)

    def check_eq(self):
        tol = self.scene.settings['3GS']['tol.angles']
        edges = list(self.diagram.edges())
        angles = self.diagram.edges_attribute('_a', keys=edges)
        amax = max(angles)
        if amax > tol:
            self.settings['_is.valid'] = False
        else:
            self.settings['_is.valid'] = True

    def clear(self):
        super(FormNetworkObject, self).clear()
        guids = []
        guids += list(self.guid_loads)
        compas_rhino.delete_objects(guids, purge=True)
        self._guid_loads = {}

    # --------------------------------------------------------------------------
    #   attributes
    # --------------------------------------------------------------------------

    # def update_attributes(self):
    #     """Update the attributes of the data structure through a Rhino dialog.

    #     Returns
    #     -------
    #     bool
    #         True if the update was successful.
    #         False otherwise.
    #     """
    #     return compas_rhino.update_settings(self.datastructure.attributes)

    # def update_nodes_attributes(self, keys, names=None):
    #     """Update the attributes of selected nodes.

    #     Parameters
    #     ----------
    #     keys : list
    #         The identifiers of the nodes of which the attributes should be updated.
    #     names : list, optional
    #         The names of the attributes that should be updated.
    #         Default is ``None``, in which case all attributes are updated.

    #     Returns
    #     -------
    #     bool
    #         True if the update was successful.
    #         False otherwise.
    #     """
    #     if keys:
    #         compas_rhino.rs.UnselectAllObjects()
    #         select_nodes(self.datastructure, keys)
    #         return network_update_node_attributes(self.datastructure, keys, names)

    # def update_edges_attributes(self, keys, names=None):
    #     """Update the attributes of selected edges.

    #     Parameters
    #     ----------
    #     keys : list
    #         The identifiers of the edges of which the attributes should be updated.
    #     names : list, optional
    #         The names of the attributes that should be updated.
    #         Default is ``None``, in which case all attributes are updated.

    #     Returns
    #     -------
    #     bool
    #         True if the update was successful.
    #         False otherwise.
    #     """
    #     if keys:
    #         compas_rhino.rs.UnselectAllObjects()
    #         select_edges(self.datastructure, keys)
    #         return network_update_edge_attributes(self.datastructure, keys, names)

    # --------------------------------------------------------------------------
    #   draw
    # --------------------------------------------------------------------------

    def draw(self):
        """Draw the objects representing the form diagram.
        """
        layer = self.settings['layer']
        self.artist.layer = layer
        self.artist.clear_layer()
        self.clear()
        if not self.visible:
            return
        self.artist.node_xyz = self.node_xyz

        # ======================================================================
        # Groups
        # ------
        # Create groups for nodes and edges.
        # These groups will be turned on/off based on the visibility settings of the diagram
        # ======================================================================

        group_nodes = "{}::nodes".format(layer)
        group_edges = "{}::edges".format(layer)
        group_angles = "{}::angles".format(layer)
        group_loads = "{}::loads".format(layer)
        group_pipes = "{}::pipes".format(layer)

        if not compas_rhino.rs.IsGroup(group_nodes):
            compas_rhino.rs.AddGroup(group_nodes)

        if not compas_rhino.rs.IsGroup(group_edges):
            compas_rhino.rs.AddGroup(group_edges)

        if not compas_rhino.rs.IsGroup(group_loads):
            compas_rhino.rs.AddGroup(group_loads)

        if not compas_rhino.rs.IsGroup(group_angles):
            compas_rhino.rs.AddGroup(group_angles)

        if not compas_rhino.rs.IsGroup(group_pipes):
            compas_rhino.rs.AddGroup(group_pipes)

        # ======================================================================
        # nodes
        # --------
        # Draw the nodes and add them to the node group.
        # ======================================================================

        if self.settings['show.nodes']:
            nodes = list(self.diagram.nodes())
            color = {node: self.settings['color.nodes'] if self.settings['_is.valid'] else self.settings['color.invalid'] for node in nodes}
            color_fixed = self.settings['color.nodes:is_fixed']
            color.update({node: color_fixed for node in self.diagram.nodes_where({'is_fixed': True}) if node in nodes})

            guids = self.artist.draw_nodes(nodes, color)
            self.guid_node = zip(guids, nodes)
            compas_rhino.rs.AddObjectsToGroup(guids, group_nodes)

            compas_rhino.rs.ShowGroup(group_nodes)
        else:
            compas_rhino.rs.HideGroup(group_nodes)

        # ======================================================================
        # Edges
        # --------
        # Draw the edges and add them to the edge group.
        # ======================================================================

        if self.settings['show.edges']:
            edges = list(self.diagram.edges())
            colors = {edge: self.settings['color.edges'] for edge in edges}
            if self.scene.settings['3GS']['show.forces']:
                colors = get_force_colors_uv(self.diagram.dual,
                                             self.diagram,
                                             gradient=True)
            colordict = {edge: colors[edge] if self.settings['_is.valid'] else self.settings['color.invalid'] for edge in edges}

            guids = self.artist.draw_edges(edges, colordict)
            self.guid_edge = zip(guids, edges)
            compas_rhino.rs.AddObjectsToGroup(guids, group_edges)

            compas_rhino.rs.ShowGroup(group_edges)
        else:
            compas_rhino.rs.HideGroup(group_edges)

        # ======================================================================
        # Angle deviations
        # --------
        # Draw angle deviatinos as FormNetwork edge labels.
        # ======================================================================

        if self.scene and self.scene.settings['3GS']['show.angles']:
            tol = self.scene.settings['3GS']['tol.angles']
            edges = list(self.diagram.edges())
            angles = self.diagram.edges_attribute('_a', keys=edges)
            amin = min(angles)
            amax = max(angles)
            if (amax - amin)**2 > 0.001**2:
                text = {}
                color = {}
                for edge, angle in zip(edges, angles):
                    if angle > tol:
                        text[edge] = "{:.0f}".format(angle)
                        color[edge] = i_to_rgb((angle - amin) / (amax - amin))
                guids = self.artist.draw_edgelabels(text, color)
                self.guid_edgelabel = zip(guids, edges)
                compas_rhino.rs.AddObjectsToGroup(guids, group_angles)
                compas_rhino.rs.ShowGroup(group_angles)
        else:
            compas_rhino.rs.HideGroup(group_angles)

        # ======================================================================
        # overlays
        # ======================================================================

        if self.settings['show.loads'] and self.settings['_is.valid']:
            scale = self.settings['scale.loads']
            guids = self.artist.draw_external_forces(gradient=False, scale=scale)
            self.guid_edge = zip(guids, self.diagram.dual.cells_on_boundaries())
            compas_rhino.rs.AddObjectsToGroup(guids, group_loads)
            compas_rhino.rs.ShowGroup(group_loads)
        else:
            compas_rhino.rs.HideGroup(group_loads)

        if self.settings['show.pipes'] and self.settings['_is.valid']:
            tol = self.settings['tol.pipes']
            edges = list(self.diagram.edges())
            colors = {edge: self.settings['color.pipes'] for edge in edges}

            if self.scene.settings['3GS']['show.forces']:
                colors = get_force_colors_uv(self.diagram.dual,
                                             self.diagram,
                                             gradient=True)
            scale = self.settings['scale.pipes']
            guids = self.artist.draw_pipes(edges, colors, scale, tol)
            self.guid_pipe = zip(guids, edges)
            compas_rhino.rs.AddObjectsToGroup(guids, group_pipes)
            compas_rhino.rs.ShowGroup(group_pipes)
        else:
            compas_rhino.rs.HideGroup(group_pipes)

        self.redraw()



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import compas

from compas.datastructures import Network

from compas.geometry import add_vectors
from compas.geometry import distance_point_point as distance
from compas.geometry import dot_vectors
from compas.geometry import cross_vectors
from compas.geometry import normalize_vector
from compas.geometry import subtract_vectors

from compas.datastructures.network.duality import network_node_find_first_neighbor

from compas.utilities import geometric_key

from compas_3gs.diagrams import EGI
from compas_3gs.diagrams import Cell

try:
    from Rhino.Geometry import Arc
    from Rhino.Geometry import Point3d
    from Rhino.Geometry import ArcCurve
    from Rhino.Geometry import Vector3d
    from Rhino.Geometry import Circle
    from Rhino.Geometry import Plane

    from Rhino.Geometry.Intersect.Intersection import CurveCurve as CCX

except ImportError:
    compas.raise_if_ironpython()


__all__ = ['egi_from_vectors',
           'cell_from_egi']


def egi_from_vectors(vectordict, origin, tol=0.001):
    """Construct an egi from a set of vectors.

    Parameters
    ----------
    vectordict : dict
        A dectionary of key-vector pairs.
    origin : list
        The coordinates of the centroid.
    tol : float, optional
        Tolerance for evaluating antipodal.

    Returns
    -------
    egi : mesh
        A mesh object representing the egi.

    Raises
    ------
    Exception
        If there are less than four vectors.

    Notes
    -----
    This algorithm is dependent on Rhinoceros objects; the adjacency arcs are implemented using Rhino.Geometry.Arc,
    and the cross-adjacencies (arc-arc intersections) are computed using Rhino.Geometry.Intersect.Intersection.CurveCurve.

    Warning
    -------
    - This algorithm does not address scenarios where multiple parallel (collinear) vectors are present.

    References
    ----------
    - Horn, B.K.P. (1984). *Extended Gaussian images*.
    - Moni, S. (1990, June). *A closed-form solution for the reconstruction
    of a convex polyhedron from its extended gaussian image.*
    - Lee, J., T. Van Mele, and P. Block (2018). *Disjointed force polyhedra.*

    """
    if len(vectordict) < 4:
        raise Exception('Four or more vectors are needed for the construction of egi.')

    egi = Network()

    # --------------------------------------------------------------------------
    #   1. add vertices from vectors
    # --------------------------------------------------------------------------
    vertex_geokeys = {}

    for vkey in vectordict:
        normal = normalize_vector(vectordict[vkey])
        vertex_xyz = add_vectors(normal, origin)
        vertex_geokeys[geometric_key(normal)] = vkey
        egi.add_vertex(x=vertex_xyz[0],
                       y=vertex_xyz[1],
                       z=vertex_xyz[2],
                       key=vkey,
                       attr_dict={'type': 'face',
                                  'normal': normal,
                                  'nbrs': []})

    # --------------------------------------------------------------------------
    #   2.  Identify main adjacencies
    # --------------------------------------------------------------------------
    vkey_pairs = set()

    for vkey in egi.vertex:
        v_crs_dict = {}

        for nbr_vkey in egi.vertex:

            if nbr_vkey is not vkey:

                n1 = egi.vertex[vkey]['normal']
                n2 = egi.vertex[nbr_vkey]['normal']

                # This checks if the normals are opposite ----------------------
                dot = dot_vectors(n1, n2)

                if dot > 1 - tol:
                    raise Exception("Coincident vectors detected.")

                elif dot > -1 + tol:

                    this_crs = cross_vectors(n1, n2)
                    unit_crs = normalize_vector(this_crs)

                    crs_gkey = geometric_key(unit_crs)

                    # Check to see if any other normals are coplanar
                    if crs_gkey not in v_crs_dict:
                        v_crs_dict[crs_gkey] = nbr_vkey

                    # If multiple arcs are coplanar, choose the closer one
                    elif crs_gkey in v_crs_dict:
                        this_dist = distance(egi.vertex_coordinates(vkey),
                                             egi.vertex_coordinates(nbr_vkey))
                        test_dist = distance(egi.vertex_coordinates(vkey),
                                             egi.vertex_coordinates(v_crs_dict[crs_gkey]))
                        if this_dist < test_dist:
                            del v_crs_dict[crs_gkey]
                            v_crs_dict[crs_gkey] = nbr_vkey

        # Add to overall connectivity dict -------------------------------------
        for crs_gkey in v_crs_dict:
            nbr_vkey = v_crs_dict[crs_gkey]
            pair = frozenset([vkey, nbr_vkey])
            vkey_pairs.add(pair)

    # --------------------------------------------------------------------------
    #   3.  Main adjacency arcs
    # --------------------------------------------------------------------------
    arcs = {}

    for pair in vkey_pairs:
        u, v = list(pair)
        arc = _draw_arc(egi.vertex[u]['normal'],
                        egi.vertex[v]['normal'],
                        origin)

        if len(arcs) == 0:
            arc_key = 0
        else:
            arc_key = max(int(x) for x in arcs.keys()) + 1
        arcs[arc_key] = {'arc': arc,
                         'vkeys': [u, v],
                         'end_vkeys': [u, v],
                         'int_vkeys': {}, }

    # --------------------------------------------------------------------------
    #   3.  arc intersections --> cross adjacencies
    # --------------------------------------------------------------------------
    arc_pairs_seen = set()
    for arckey_1 in arcs:
        for arckey_2 in arcs:
            if arckey_1 != arckey_2:
                arc_pair = frozenset([arckey_1, arckey_2])
                if arc_pair not in arc_pairs_seen:
                    arc_1 = arcs[arckey_1]['arc']
                    arc_2 = arcs[arckey_2]['arc']
                    intersection = _curve_curve_intx(arc_1, arc_2)
                    if intersection:
                        new_vkey = max(int(vkey) for vkey in egi.vertex.keys()) + 1
                        new_normal = subtract_vectors(intersection, origin)
                        new_normal = normalize_vector(new_normal)
                        new_vertex_geokey = geometric_key(new_normal, precision='3f')

                        # if intersection is not an endpoint -------------------
                        if new_vertex_geokey not in vertex_geokeys.keys():
                            vertex_geokeys[new_vertex_geokey] = new_vkey
                            egi.add_vertex(x=intersection[0],
                                           y=intersection[1],
                                           z=intersection[2],
                                           key=new_vkey,
                                           attr_dict={'type': 'zero',
                                                      'normal': new_normal,
                                                      'magnitude': 0,
                                                      'nbrs': []})
                            arcs[arckey_1]['vkeys'].append(new_vkey)
                            arcs[arckey_2]['vkeys'].append(new_vkey)
                            arcs[arckey_1]['int_vkeys'][new_vkey] = arckey_2
                            arcs[arckey_2]['int_vkeys'][new_vkey] = arckey_1

                        # if intersection already exists -----------------------
                        elif new_vertex_geokey in vertex_geokeys.keys():
                            vkey = vertex_geokeys[new_vertex_geokey]
                            if vkey not in arcs[arckey_1]['vkeys']:
                                arcs[arckey_1]['vkeys'].append(vkey)
                                arcs[arckey_1]['int_vkeys'][vkey] = arckey_2
                            if vkey not in arcs[arckey_2]['vkeys']:
                                arcs[arckey_2]['vkeys'].append(vkey)
                                arcs[arckey_2]['int_vkeys'][vkey] = arckey_1
                        arc_pairs_seen.add(arc_pair)

    # --------------------------------------------------------------------------
    #   5.  Reorder vertices along each arc and add edges to EGI network
    # --------------------------------------------------------------------------
    for arckey in arcs:
        vkeys = arcs[arckey]['vkeys']
        if len(vkeys) > 2:
            pt_list = [egi.vertex_coordinates(key) for key in vkeys]
            arcs[arckey]['vkeys'] = _reorder_pts_on_arc(pt_list,
                                                        arcs[arckey]['vkeys'],
                                                        arcs[arckey]['arc'])[1]
        #     edge_type = 'cross'
        # else:
        #     edge_type = 'main'
        for i in range(len(arcs[arckey]['vkeys']) - 1):
            vkey_1 = arcs[arckey]['vkeys'][i]
            vkey_2 = arcs[arckey]['vkeys'][i + 1]
            egi.vertex[vkey_1]['nbrs'] += [vkey_2]
            egi.vertex[vkey_2]['nbrs'] += [vkey_1]
            egi.add_edge(vkey_1, vkey_2)

    # --------------------------------------------------------------------------
    #   6.  For each vertex, sort nbrs in ccw order
    # --------------------------------------------------------------------------
    _egi_sort_v_nbrs(egi)

    # --------------------------------------------------------------------------
    #   7.  Add EGI Network faces
    # --------------------------------------------------------------------------
    egi_mesh = EGI()
    for vkey in egi.vertex:
        egi_mesh.vertex[vkey] = egi.vertex[vkey]

    egi_mesh.attributes['name'] = 'egi'
    egi_mesh.attributes['origin'] = list(origin)

    _egi_find_faces(egi, egi_mesh)

    return egi_mesh


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   from egi to unit polyhedron
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************

def cell_from_egi(egi):
    """Construct a cell from an egi.

    Parameters
    ----------
    egi
        A mesh object representing the egi.

    Returns
    -------
    cell
        A mesh object representing the unit polyhedron.

    """
    cell = Cell()
    cell.name = 'cell'

    for fkey in egi.face:
        x, y, z = egi.face_center(fkey)
        cell.add_vertex(key=fkey, x=x, y=y, z=z)

    for vkey in egi.vertex:
        cell_face = egi.vertex_faces(vkey, ordered=True)
        cell.add_face(cell_face[::-1], fkey=vkey)

        cell.facedata[vkey]['type'] = egi.vertex[vkey]['type']

    return cell


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   helpers
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def _draw_arc(normal_1, normal_2, origin):
    mid_pt = normalize_vector(add_vectors(normal_1, normal_2))
    arc = Arc(Point3d(*[sum(axis) for axis in zip(normal_1, origin)]),
              Point3d(*[sum(axis) for axis in zip(mid_pt, origin)]),
              Point3d(*[sum(axis) for axis in zip(normal_2, origin)]))
    arc_as_curve = ArcCurve(arc)
    return arc_as_curve


def _reorder_pts_on_arc(pt_list, pt_key_list, arc_curve):
    # all points should be on the arc...
    dist_list = []
    sp = arc_curve.PointAtStart
    for pt in pt_list:
        dist_list.append(distance(sp, pt))
    ordered_pt_list = [x for (y, x) in sorted(zip(dist_list, pt_list))]
    ordered_pt_key_list = [x for (y, x) in sorted(zip(dist_list, pt_key_list))]
    return ordered_pt_list, ordered_pt_key_list


def _curve_curve_intx(curve_1, curve_2):
    intersection_tolerance = 0.01
    overlap_tolerance = 0.0
    intersection = CCX(curve_1,
                       curve_2,
                       intersection_tolerance,
                       overlap_tolerance)
    if not intersection:
        return None
    for instance in intersection:
        return instance.PointA


def _egi_sort_v_nbrs(egi):
    """ By default, the sorting should be ccw, since the circle is typically drawn
    ccw around the local plane's z-axis...
    """
    xyz = dict((key, [attr[_] for _ in 'xyz']) for key, attr in egi.vertices(True))
    for vkey in egi.vertex:
        nbrs = egi.vertex[vkey]['nbrs']
        plane = Plane(Point3d(*xyz[vkey]),
                      Vector3d(*[axis for axis in egi.vertex[vkey]['normal']]))
        circle = Circle(plane, 1)
        p_list = []
        for nbr_vkey in nbrs:
            boolean, parameter = ArcCurve(circle).ClosestPoint(Point3d(*xyz[nbr_vkey]))
            p_list.append(parameter)
        sorted_nbrs = [key for (param, key) in sorted(zip(p_list, nbrs))]
        egi.vertex[vkey]['sorted_nbrs'] = sorted_nbrs


def _egi_find_edge_face(u, v, egi):
    """ same as duality.algorithms.find_edge_faces... using 'sorted_nbrs' instead
    """
    cycle = [u]
    while True:
        cycle.append(v)
        nbrs = egi.vertex[v]['sorted_nbrs']
        nbr = nbrs[nbrs.index(u) - 1]
        u, v = v, nbr
        if v == cycle[0]:
            cycle.append(v)
            break
    return cycle


def _egi_find_faces(egi, egi_mesh):
    """ Modified, and simplified version of duality.algorithms.find_network_faces...
    since there are no leaves or open faces in a egi network.
    """
    egi_mesh.halfedge = {key: {} for key in egi.vertices()}
    for u, v in egi.edges():
        egi_mesh.halfedge[u][v] = None
        egi_mesh.halfedge[v][u] = None
    u = sorted(egi.vertices(True), key=lambda x: (x[1]['y'], x[1]['x']))[0][0]
    v = network_node_find_first_neighbor(egi, u)

    egi_mesh.add_face(_egi_find_edge_face(u, v, egi))

    for u, v in egi.edges():

        if egi_mesh.halfedge[u][v] is None:

            egi_mesh.add_face(_egi_find_edge_face(u, v, egi))
        if egi_mesh.halfedge[v][u] is None:
            egi_mesh.add_face(_egi_find_edge_face(v, u, egi))

    return egi_mesh


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import json

import scriptcontext as sc

import compas_rhino

from compas.utilities import DataDecoder


from compas_3gs.diagrams import ForceVolMesh
from compas_3gs.diagrams import FormNetwork

__commandname__ = "TGS__session_load"


def RunCommand(is_interactive):

    if '3GS' not in sc.sticky:
        compas_rhino.display_message('3GS has not been initialised yet.')
        return

    system = sc.sticky['3GS']['system']
    scene = sc.sticky['3GS']['scene']

    filepath = compas_rhino.select_file(folder=system['session.dirname'], filter=system['session.extension'])

    if not filepath:
        return
    if not os.path.exists(filepath):
        return
    if not os.path.isfile(filepath):
        return
    if not filepath.endswith(".{}".format(system['session.extension'])):
        return

    dirname, basename = os.path.split(filepath)
    filename, extension = os.path.splitext(basename)

    system['session.dirname'] = dirname
    system['session.filename'] = filename

    with open(filepath, "r") as f:
        session = json.load(f, cls=DataDecoder)

    if not session['data']['force']:
        compas_rhino.display_message('The session file has no force diagram.')

    forcediagram = ForceVolMesh.from_data(session['data']['force'])

    force_id = scene.add_forcevolmesh(forcediagram, name='force', layer='3GS::ForceDiagram')
    force = scene.find(force_id)

    if 'settings' in session['scene']['force']:
        force.settings.update(session['scene']['force']['settings'])

    if 'anchor' in session['scene']['force']:
        force.anchor = session['scene']['force']['anchor']

    if 'location' in session['scene']['force']:
        force.location = session['scene']['force']['location']

    if 'scale' in session['scene']['form']:
        force.scale = session['scene']['force']['scale']

    if session['data']['form']:

        formdiagram = FormNetwork.from_data(session['data']['form'])

        form_id = scene.add_formnetwork(formdiagram, name='form', layer='3GS::FormDiagram')
        form = scene.find(form_id)

        forcediagram.primal = formdiagram
        formdiagram.dual = forcediagram

        if 'settings' in session['scene']['form']:
            form.settings.update(session['scene']['form']['settings'])

        if 'anchor' in session['scene']['form']:
            form.anchor = session['scene']['form']['anchor']

        if 'location' in session['scene']['form']:
            form.location = session['scene']['form']['location']

        if 'scale' in session['scene']['form']:
            form.scale = session['scene']['form']['scale']

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    RunCommand(True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.geometry import cross_vectors
from compas.geometry import subtract_vectors
from compas.geometry import normalize_vector
from compas.geometry import dot_vectors
from compas.geometry import length_vector
from compas.geometry import add_vectors
from compas.geometry import scale_vector
from compas.geometry import centroid_points_weighted
from compas.geometry import centroid_points
from compas.geometry import midpoint_point_point
from compas.geometry import bestfit_plane
from compas.geometry import distance_point_point
from compas.geometry import intersection_line_line
from compas.geometry import is_point_on_segment
from compas.geometry import project_point_plane


__all__ = ['resultant_vector',

           'polygon_normal_oriented',
           'polygon_area_oriented',
           'polygon_area_footprint',
           'scale_polygon',

           'cell_face_flatness',
           'cell_face_areaness',

           'volmesh_face_flatness',
           'volmesh_face_areaness',

           'datastructure_centroid']


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   vectors
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def resultant_vector(vectors, locations):
    """Computes the resultant vector of a group of vectors.

    Parameters
    ----------
    vectors: dictionary
        vector_key-XYZ (vector components) pairs

    locations: dictionary
        vector_key-XYZ (vector locations) pairs

    Returns
    -------
    vector: tuple
        XYZ components of the resultant vector

    point: tuple
        XYZ coordinates of the location of the resultant vector

    Examples
    --------
    >>> resultant_vector({0 : [ 0.97 , 0.83 , 0.83 ],
                          1 : [ 0.65 , 0.89 , 0.34 ],
                          2 : [ 0.68 , 0.63 , 0.97 ]},
                         {0 : [ 0.97 , 0.83 , 0.83 ],
                          1 : [ 0.65 , 0.89 , 0.34 ],
                          2 : [ 0.68 , 0.63 , 0.97 ]})
    [2.3000000000000003, 2.35, 2.1399999999999997]
    [0.78129727344020572, 0.78043477156745522, 0.7360930822393742],

    """
    points = []
    weights = []

    for vkey in vectors:
        points.append(vectors[vkey])
        weights.append(length_vector(vectors[vkey]))

    resultant_xyz = centroid_points_weighted(points, weights)
    x, y, z = zip(*vectors.values())
    resultant_vector = [sum(x), sum(y), sum(z)]

    return resultant_vector, resultant_xyz


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   polygons
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def polygon_normal_oriented(polygon, unitized=True):
    """Compute the oriented normal of any closed polygon (can be convex, concave or complex).

    Parameters
    ----------
    polygon : sequence
        The XYZ coordinates of the vertices/corners of the polygon.
        The vertices are assumed to be in order.
        The polygon is assumed to be closed:
        the first and last vertex in the sequence should not be the same.

    Returns
    -------
    list
        The weighted or unitized normal vector of the polygon.

    """
    p = len(polygon)
    assert p > 2, "At least three points required"
    w = centroid_points(polygon)
    normal_sum = (0, 0, 0)

    for i in range(-1, len(polygon) - 1):
        u = polygon[i]
        v = polygon[i + 1]
        uv = subtract_vectors(v, u)
        vw = subtract_vectors(w, v)
        normal = scale_vector(cross_vectors(uv, vw), 0.5)
        normal_sum = add_vectors(normal_sum, normal)

    if not unitized:
        return normal_sum

    return normalize_vector(normal_sum)


def polygon_area_oriented(polygon):
    """Compute the oriented area of a polygon (can be convex or concave).

    Parameters
    ----------
    polygon : sequence
        The XYZ coordinates of the vertices/corners of the polygon.
        The vertices are assumed to be in order.
        The polygon is assumed to be closed:
        the first and last vertex in the sequence should not be the same.

    Returns
    -------
    float
        The area of the polygon.

    """
    return length_vector(polygon_normal_oriented(polygon, unitized=False))


def polygon_area_footprint(polygon):
    """Compute the non-oriented area of a polygon (can be convex or concave).

    Parameters
    ----------
    polygon : list of lists
        A list of polygon point coordinates.

    Returns
    -------
    float
        The non-oriented area of the polygon.

    """
    area = 0
    w = centroid_points(polygon)

    for i in range(-1, len(polygon) - 1):
        u = polygon[i]
        v = polygon[i + 1]
        uv = subtract_vectors(v, u)
        vw = subtract_vectors(w, v)
        normal = scale_vector(cross_vectors(uv, vw), 0.5)
        area += length_vector(normal)

    return area


def scale_polygon(polygon, scale):
    """Scale a polygon.

    Parameters
    ----------
    polygon : list of lists
        A list of polygon point coordinates.

    Returns
    -------
    list of lists
        Reordered polygon point coordinates.

    """
    center = centroid_points(polygon)

    scaled_polygon = []

    for pt in polygon:
        vector = subtract_vectors(pt, center)
        new_pt = add_vectors(center, scale_vector(vector, scale))
        scaled_polygon.append(new_pt)

    return scaled_polygon


def untangle_polygon(polygon):
    """Untangle a polygon.

    Parameters
    ----------
    polygon : list of lists
        A list of polygon point coordinates.

    Returns
    -------
    list of lists
        Reordered polygon point coordinates.

    """
    def _cross_edges(edge1, edge2):
        a, b = edge1
        c, d = edge2
        edge1_vec = normalize_vector(subtract_vectors(b, a))
        edge2_vec = normalize_vector(subtract_vectors(d, c))
        cross = cross_vectors(edge1_vec, edge2_vec)
        return cross

    xyz = {i: polygon[i] for i in range(len(polygon))}
    vkeys = range(len(polygon))

    count = 20
    while count:
        count -= 1
        moved = []

        for i in range(-1, len(polygon) - 1):
            t = xyz[vkeys[i - 2]]
            u = xyz[vkeys[i - 1]]
            v = xyz[vkeys[i]]
            w = xyz[vkeys[i + 1]]

            u_cross = _cross_edges((t, u), (u, v))
            v_cross = _cross_edges((u, v), (v, w))

            dot = dot_vectors(u_cross, v_cross)

            if dot < 0:
                midpt = midpoint_point_point(u, w)
                vec = scale_vector(subtract_vectors(midpt, v), 1.5)

                xyz[vkeys[i]] = add_vectors(v, vec)
                moved.append([vkeys[i]])

        if len(moved) == 0:
            break

    return [xyz[i] for i in vkeys]


def polygon_flatness(polygon):
    """Comput the flatness of a polygon.

    Parameters
    ----------
    polygon : list of lists
        A list of polygon point coordinates.

    Returns
    -------
    float
        The flatness.

    Note
    ----
    compas.geometry.mesh_flatness function currently only works for quadrilateral faces.
    This function uses the distance between each face vertex and its projected point on the best-fit plane of the face as the flatness metric.

    """
    deviation = 0

    plane = bestfit_plane(polygon)

    for pt in polygon:
        pt_proj = project_point_plane(pt, plane)
        dev = distance_point_point(pt, pt_proj)
        if dev > deviation:
            deviation = dev

    return deviation


def is_polygon_self_intersecting(polygon):
    """Computes if as polygon is self intersecting in plane, or self overlapping in space.

    Parameters
    ----------
    polygon : list of lists
        A list of polygon point coordinates.

    Returns
    -------
    bool_1
        ``True`` if self overlapping.
        ``False`` otherwise.
    bool_2
        ``True`` if self intersecting.
        ``False`` otherwise.

    """
    edges = []
    for i in range(-1, len(polygon) - 1):
        edges.append((i, i + 1))

    for u1, v1 in edges:
        for u2, v2 in edges:
            if u1 == u2 or v1 == v2 or u1 == v2 or u2 == v1:
                continue
            else:
                a = polygon[u1]
                b = polygon[v1]
                c = polygon[u2]
                d = polygon[v2]

                int_1, int_2 = intersection_line_line((a, b), (c, d))

                if int_1 or int_2:
                    if distance_point_point(int_1, int_2) > 0:
                        overlapping = True
                    if is_point_on_segment(int_1, (a, b)) or is_point_on_segment(int_2, (c, d)):
                        intersecting = True

    return overlapping, intersecting

# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   mesh
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def cell_face_flatness(cell):
    """Compute the flatness of every face of a mesh.

    Parameters
    ----------
    cell : Mesh
        Cell as a mesh object.

    Returns
    -------
    dict
        A dictionary of flatness deviation for every face of the mesh.

    """

    flatness_dict = {fkey: 0 for fkey in cell.face}

    for fkey in cell.face:
        f_vkeys = cell.face_vertices(fkey)
        f_pts = [cell.vertex_coordinates(vkey) for vkey in f_vkeys]
        dev = polygon_flatness(f_pts)

        flatness_dict[fkey] = dev

    return flatness_dict


def cell_face_areaness(cell, target_areas):
    """Compute the areaness of every face of a mesh.

    Parameters
    ----------
    cell : Mesh
        Cell as a mesh object.
    target_areas : dict
        Dictionary of target areas.

    Returns
    -------
    dict
        A dictionary of area deviation for each face of the mesh.

    """
    areaness_dict = {}

    for fkey in target_areas:
        area = cell.face_area(fkey)
        areaness_dict[fkey] = abs(target_areas[fkey] - area)

    return areaness_dict


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def volmesh_face_flatness(volmesh):
    """Compute the flatness of every face of a volmesh.

    Parameters
    ----------
    volmesh : volmesh object

    Returns
    -------
    dict
        A dictionary of flatness deviation for every face of the volmesh.

    """
    flatness_dict = {fkey: 0 for fkey in volmesh.faces()}

    for fkey in volmesh.faces():
        f_vkeys = volmesh.halfface_vertices(fkey)
        f_pts = [volmesh.vertex_coordinates(vkey) for vkey in f_vkeys]
        dev = polygon_flatness(f_pts)

        flatness_dict[fkey] = dev

    return flatness_dict


def volmesh_face_areaness(volmesh, target_areas):
    """Compute the areaness of every face of a volmesh.

    Parameters
    ----------
    volmesh : volmesh object

    Returns
    -------
    dict
        A dictionary of area deviation for each face of the volmesh.

    """
    areaness_dict = {}

    for hfkey in target_areas:
        area = volmesh.halfface_area(hfkey)
        areaness_dict[hfkey] = abs(target_areas[hfkey] - area)

    return areaness_dict


def _get_current_volmesh_normals(volmesh):
    normal_dict = {}
    for hfkey in volmesh.halfface:
        center = volmesh.halfface_center(hfkey)
        normal = volmesh.halfface_normal(hfkey)
        normal_dict[hfkey] = {'normal': normal, 'center': center}
    return normal_dict


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   datastructures
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def datastructure_centroid(datastructure):
    """Compute the centroid of the datastructure.

    Parameters
    ----------
    datastructure
        A network, mesh or volmesh object.

    Returns
    -------
    list
        The coordinates of the centroid.

    """
    points = [datastructure.vertex_coordinates(vkey) for vkey in datastructure.vertices()]

    return centroid_points(points)


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import compas
import compas_rhino

from compas.geometry import add_vectors
from compas.geometry import scale_vector
from compas.geometry import subtract_vectors
from compas.geometry import normalize_vector
from compas.geometry import length_vector

from compas.utilities import i_to_green
from compas.utilities import color_to_colordict

from compas_3gs.algorithms import volmesh_ud

from compas_3gs.utilities import pair_uv_to_hf

from compas_3gs.utilities import get_index_colordict
from compas_3gs.utilities import valuedict_to_colordict
from compas_3gs.utilities import get_force_mags
from compas_3gs.utilities import get_force_colors_uv
from compas_3gs.utilities import get_force_colors_hf

try:
    import rhinoscriptsyntax as rs
    import scriptcontext as sc

    from System.Drawing.Color import FromArgb

    from Rhino.Geometry import Point3d
    from Rhino.Geometry import Arc
    from Rhino.Geometry import ArcCurve
    from Rhino.Geometry import Sphere
    from Rhino.Geometry import Vector3d
    from Rhino.Geometry import Plane
    from Rhino.Geometry import Brep

    from Rhino.DocObjects.ObjectColorSource import ColorFromObject
    from Rhino.DocObjects.ObjectDecoration import EndArrowhead

    find_object = sc.doc.Objects.Find
    find_layer_by_fullpath = sc.doc.Layers.FindByFullPath
    add_arc = sc.doc.Objects.AddArc
    add_brep = sc.doc.Objects.AddBrep
    add_curve = sc.doc.Objects.AddCurve

except ImportError:
    compas.raise_if_ironpython()


__all__ = ['draw_vertex_normal',
           'draw_vertex_fixities',

           'draw_cell_force_vectors',
           'draw_cell_labels',
           'clear_cell_labels',

           'draw_network_external_forces',
           'draw_network_pipes',
           'draw_network_internal_forces',

           'draw_volmesh_cells',
           'draw_corresponding_elements',
           'draw_compression_tension',
           'draw_directed_hf_and_uv',
           'draw_volmesh_face_normals',
           'draw_volmesh_UD',

           'draw_egi_arcs',

           'bake_cells_as_polysurfaces']


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   vertex
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def draw_vertex_normal(datastructure, vkeys=None):
    pass


def draw_vertex_fixities(diagram):

    text_dict = {}
    color_dict = {}
    for vertex in diagram.vertices():

        text = ''
        r = 0
        g = 0
        b = 0
        if diagram.vertex_attribute(vertex, 'x_fix') is True:
            text += 'x'
            r = 255
        if diagram.vertex_attribute(vertex, 'y_fix') is True:
            text += 'y'
            g = 255
        if diagram.vertex_attribute(vertex, 'z_fix') is True:
            text += 'z'
            b = 255

        if text != '':
            text_dict[vertex] = text
            color_dict[vertex] = (r, g, b)

    diagram.draw()
    guids = diagram.draw_vertexlabels(text=text_dict, color=color_dict)
    return guids


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   face
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def draw_face_normal(datastructure, fkeys=None):
    pass


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   cells
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


# def draw_cell(volmesh, ckeys=None):
#     """Draw the specified cells of a volmesh.

#     Parameters
#     ----------
#     ckeys: list, optional
#         The keys of specific cells that should be drawn.

#     """
#     if not ckeys:
#         return
#     hfkeys = []
#     for ckey in ckeys:
#         hfkeys += volmesh.cell_halffaces(ckey)
#     volmesh.draw_faces(keys=hfkeys)


def draw_cell_force_vectors(volmesh, ckey):
    """Draw the force vectors of a single cell.
    """
    center = volmesh.cell_centroid(ckey)
    lines = []

    for hfkey in volmesh.halfface:
        normal = volmesh.halfface_normal(hfkey, unitized=False)

        lines.append({
            'start': center,
            'end': add_vectors(center, normal),
            'arrow': 'end',
            'color': (0, 255, 0),
            'name': 'hfkey.{}'.format(hfkey)})

    guids = compas_rhino.draw_lines(lines)

    return guids


def draw_cell_labels(volmesh, ckeys=None, text=None, color=None):
    """Draw cell labels.

    Parameters
    ----------
    text : dict
        A dictionary of ckey-text pairs.
        The default value is ``None``, in which case every cell will be labelled with its key.
    colors : str, tuple, dict
        The color specification for the cells.

    """
    if ckeys is None:
        ckeys = volmesh.cell

    # 1. get ckeys to label ----------------------------------------------------
    if text is None:
        textdict = {ckey: str(ckey) for ckey in ckeys}
    elif isinstance(text, dict):
        textdict = text
    else:
        raise NotImplementedError

    # 2. get colors ------------------------------------------------------------
    colordict = get_index_colordict(volmesh.cell)
    if color:
        colordict = color_to_colordict(color,
                                       textdict.keys(),
                                       default=(0, 0, 0),
                                       colorformat='rgb',
                                       normalize=False)

    # 3. labels ----------------------------------------------------------------
    labels = []
    for ckey in textdict:
        labels.append({
            'pos': volmesh.cell_centroid(ckey),
            'name': '{0}.cell.label.{1}'.format(volmesh.name, ckey),
            'color': colordict[ckey],
            'text': textdict[ckey]})

    # 4. draw ------------------------------------------------------------------
    guids = compas_rhino.draw_labels(labels,
                                     layer=volmesh.layer,
                                     clear=False,
                                     redraw=False)
    return guids


def clear_cell_labels(volmesh, ckeys=None):
    """Clear cell labels.

    Parameters
    ----------
    keys : list, optional
        The keys of a specific set of cell labels that should be cleared.
        Default is to clear all cell labels.

    """
    if not ckeys:
        name = '{}.cell.label.*'.format(volmesh.name)
        guids = compas_rhino.get_objects(name=name)
    else:
        guids = []
        for key in ckeys:
            name = '*.cell.label.{}'.format(key)
            guid = compas_rhino.get_object(name=name)
            guids.append(guid)
    guids = compas_rhino.delete_objects(guids)
    return guids


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def draw_volmesh_cells(volmesh,
                       ckeys=[],
                       color=None):

    # 2. get hf_colors ---------------------------------------------------------
    hf_colordict = color_to_colordict(color,
                                      keys=volmesh.halfface.keys(),
                                      default=(0, 0, 0),
                                      colorformat='rgb',
                                      normalize=False)

    hfkeys = []

    for ckey in ckeys:
        halffaces = volmesh.cell_halffaces(ckey)
        for hfkey in halffaces:
            hfkeys.append(hfkey)

    guids = volmesh.draw_faces(faces=hfkeys, color=hf_colordict)
    return guids


def draw_volmesh_face_normals(volmesh,
                              hfkeys=None,
                              color=None,
                              scale=1):
    """Draws the face normals of the volmesh.

    Parameters
    ----------
    volmesh : volmesh object
        A volmesh datastructure representing a polyhedral force diagram.
    hfkeys : list, optional
        Specific hfkeys to display the normals for.
    hf_c_dict : dictionary, optional
        An optinoal dictionary of hfkey-color pairs.
    scale : float, optional
        Scale factor for the face normal vectors.

    Note
    ----
        By default, normals of volmesh.faces() are drawn. For the interior cell-to-cell interfaces, the normal will be drawn for only one of the two halffaces at that interface.

    """

    # 1. evaluate hfkeys -------------------------------------------------------
    hfkeys = hfkeys or volmesh.faces()

    # 2. get hf_colors ---------------------------------------------------------
    hf_colordict = color_to_colordict(color,
                                      hfkeys,
                                      default=(0, 255, 0),
                                      colorformat='rgb',
                                      normalize=False)

    # 2. normals to draw -------------------------------------------------------
    f_normals = []
    for hfkey in hfkeys:
        center = volmesh.face_center(hfkey)
        normal = scale_vector(volmesh.halfface_normal(hfkey), scale)
        name = '{}.edge.f_normal.{}'.format(volmesh.attributes['name'], hfkey)
        f_normals.append({
            'start': center,
            'end': add_vectors(center, normal),
            'arrow': 'end',
            'color': hf_colordict[hfkey],
            'name': name})

    # 3. draw ------------------------------------------------------------------
    guids = compas_rhino.draw_lines(f_normals,
                                    layer=volmesh.layer,
                                    clear=False,
                                    redraw=False)
    return guids


def draw_volmesh_UD(volmesh,
                    network,
                    color=None,
                    draw_prisms=True,
                    scale=0.5):
    """Draw the unified diagram of a volmesh.
    """

    hfkeys = volmesh.halfface.keys()

    # 1. get colors ------------------------------------------------------------
    hf_color = (0, 0, 0)
    prism_c = (155, 225, 255)  # color for compression prisms
    prism_t = (255, 180, 180)  # color for tension prisms

    hf_colordict = color_to_colordict(color,
                                      hfkeys,
                                      default=hf_color,
                                      colorformat='rgb',
                                      normalize=False)

    # 2. compute unified diagram geometries ------------------------------------
    halffaces, prism_faces = volmesh_ud(volmesh, network, scale=scale)

    # 3. halffaces and prisms --------------------------------------------------
    faces = []
    for hfkey in hfkeys:
        vkeys = volmesh.halfface[hfkey]
        hf_xyz = [halffaces[hfkey][i] for i in vkeys]
        name = '{}.face.ud.{}'.format(volmesh.name, hfkey)
        faces.append({'points': hf_xyz,
                      'name': name,
                      'color': hf_colordict[hfkey]})

    if draw_prisms:
        forces = get_force_mags(volmesh, network)
        for uv in prism_faces:
            name = '{}.face.ud.prism.{}'.format(volmesh.name, uv)
            color = prism_c
            if forces[uv] > 0:
                color = prism_t
            for face in prism_faces[uv]:
                faces.append({'points': face,
                              'name': name,
                              'color': color})

    # 4. draw ------------------------------------------------------------------
    guids = compas_rhino.draw_faces(faces,
                                    layer=volmesh.layer,
                                    clear=False,
                                    redraw=False)

    return guids

# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   network
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def draw_network_external_forces(volmesh,
                                 network,
                                 gradient=True,
                                 scale=1.0):
    """Draws the boundary forces on the form diagram.

    Parameters
    ----------
    volmesh : volmesh object
        A volmesh datastructure representing a polyhedral force diagram.
    network : network object
        A network datastructure representing a polyhedral form diagram.
    scale : float, optional
        Drawing scale factor for the boundary forces.

    """

    load_vectors = {}
    load_mags = {}

    for cell in volmesh.cells_on_boundaries():
        vectors = []
        for face in volmesh.cell_faces(cell):
            if volmesh.is_halfface_on_boundary(face):
                vectors.append(volmesh.halfface_normal(face, unitized=False))

        vectors = zip(*vectors)
        resultant = [sum(vector) for vector in vectors]
        load_vectors[cell] = resultant
        load_mags[cell] = length_vector(resultant)

    # get green gradient
    colordict = valuedict_to_colordict(load_mags, color_scheme=i_to_green)

    lines = []
    for vertex in load_vectors:
        vector = scale_vector(load_vectors[vertex], scale)
        sp = network.vertex_coordinates(vertex)
        ep = add_vectors(sp, vector)
        if volmesh.attributes['convention'] == -1:
            sp, ep = ep, sp
        color = (0, 100, 0)
        if gradient:
            color = colordict[vertex]
        lines.append({
            'start': sp,
            'end': ep,
            'color': color,
            'arrow': 'end',
            'name': '{}.load.at.vertex.{}'.format(network.name, vertex)})

    # 5. draw ------------------------------------------------------------------
    return compas_rhino.draw_lines(lines,
                                   layer=network.layer,
                                   clear=False,
                                   redraw=False)

    # # # 2. get boundary hfs ------------------------------------------------------
    # hfkeys = volmesh.halffaces_on_boundaries()

    # # # 1. delete any existing boundary forces -----------------------------------
    # # name = '*.edge.loads.*'
    # # delete_objects(compas_rhino.get_objects(name=name))

    # # # evaluate specified vertices
    # # if vkeys:
    # #     temp_hfs = []
    # #     for vkey in vkeys:
    # #         cell_hfs = volmesh.cell_halffaces(vkey)
    # #         for cell_hf in cell_hfs:
    # #             if cell_hf in hfkeys:
    # #                 temp_hfs.append(cell_hf)
    # #     if not temp_hfs:
    # #         return
    # #     hfkeys = temp_hfs

    # # 3. get hf areas and colors -----------------------------------------------
    # hf_areas = {hfkey: volmesh.halfface_area(hfkey) for hfkey in hfkeys}

    # hf_c_dict = {}
    # if gradient:
    #     hf_c_dict = valuedict_to_colordict(hf_areas, color_scheme=i_to_green)

    # colordict = color_to_colordict(hf_c_dict,
    #                                hfkeys,
    #                                default=(0, 255, 0),
    #                                colorformat='rgb',
    #                                normalize=False)

    # # 4. forces and labels to draw ---------------------------------------------
    # lines = []
    # for hfkey in hfkeys:
    #     ckey = volmesh.halfface_cell(hfkey)
    #     normal = volmesh.halfface_normal(hfkey, unitized=False)
    #     area = hf_areas[hfkey]
    #     vector = scale_vector(normal, scale)

    #     sp = network.vertex_coordinates(ckey)
    #     ep = add_vectors(sp, vector)
    #     if volmesh.attributes['convention'] == -1:
    #         sp, ep = ep, sp

    #     color = colordict[hfkey]
    #     lines.append({
    #         'start': sp,
    #         'end': ep,
    #         'color': color,
    #         'arrow': 'end',
    #         'name': '{}.edge.loads.{}.{}'.format(network.name, hfkey, area)})

    # # 5. draw ------------------------------------------------------------------
    # guids = compas_rhino.draw_lines(lines,
    #                                 layer=network.layer,
    #                                 clear=False,
    #                                 redraw=False)

    # return guids


def draw_network_internal_forces(volmesh,
                                 network,
                                 vkeys=[],
                                 gradient=True,
                                 scale=1.0):

    # 1. get vkeys -------------------------------------------------------------
    vkeys = vkeys or network.nodes()

    # 2. colors and force magnitudes -------------------------------------------
    uv_f_dict = get_force_mags(volmesh, network)
    uv_c_dict = get_force_colors_uv(volmesh, network, gradient=gradient)

    # 3. vectors to draw -------------------------------------------------------
    arrows = []

    for u in vkeys:
        for v in network.neighbors(u):

            vec = network.edge_direction(u, v)
            u_xyz = network.node_coordinates(u)

            if (u, v) in uv_f_dict:
                force = uv_f_dict[(u, v)]
                color = uv_c_dict[(u, v)]

            else:
                force = uv_f_dict[(v, u)]
                color = uv_c_dict[(v, u)]

            v_xyz = add_vectors(u_xyz, scale_vector(vec, abs(force) * scale))

            sp = u_xyz
            ep = v_xyz
            if force < 0:  # if compression
                sp = v_xyz
                ep = u_xyz

            arrows.append({
                'start': sp,
                'end': ep,
                'color': color,
                'arrow': 'end',
                'layer': network.layer,
                'name': '{}.edge.force.{}-{}'.format(network.name, u, v)})

    # 4. draw force vectors ----------------------------------------------------
    guids = compas_rhino.draw_lines(arrows)

    return guids


def draw_network_pipes(volmesh,
                       network,
                       color=None,
                       scale=1.0):

    colordict = color_to_colordict(color,
                                   [uv for uv in network.edges()],
                                   default=(0, 0, 0),
                                   colorformat='rgb',
                                   normalize=False)

    # 2. map uv to hf ----------------------------------------------------------
    uv_hf_dict = pair_uv_to_hf(network, volmesh)

    # 3. pipes to draw ---------------------------------------------------------
    cylinders = []
    for uv in network.edges():
        cylinders.append({
            'start': network.vertex_coordinates(uv[0]),
            'end': network.vertex_coordinates(uv[1]),
            'radius': volmesh.halfface_area(uv_hf_dict[uv]) * scale,
            'color': colordict[uv],
            'layer': network.layer,
            'name': '{}.edge.pipes.{}'.format(network.name, uv)})

    # 4. draw ------------------------------------------------------------------
    guids = compas_rhino.draw_cylinders(cylinders, cap=True)
    return guids

# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   both diagrams
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def draw_corresponding_elements(volmesh,
                                datastructure,
                                edges=[]):

    if not edges:
        edges = list(datastructure.edges())

    uv_c_dict = get_index_colordict(edges)
    hf_c_dict = get_force_colors_hf(volmesh, datastructure, uv_c_dict=uv_c_dict)

    faces_to_draw = [pair_uv_to_hf(datastructure, volmesh)[uv] for uv in edges]

    volmesh.clear()
    volmesh.draw_edges()
    volmesh.draw_faces(faces=faces_to_draw, color=hf_c_dict)

    datastructure.clear()
    datastructure.draw_edges(color=uv_c_dict)


def draw_compression_tension(volmesh,
                             network,
                             gradient=False,
                             label=False):

    hfkeys = pair_uv_to_hf(network, volmesh).values()

    uv_c_dict = get_force_colors_uv(volmesh, network, gradient=gradient)
    hf_c_dict = get_force_colors_hf(volmesh, network, uv_c_dict=uv_c_dict)

    volmesh.clear()
    volmesh.draw_edges()
    volmesh.draw_faces(faces=hfkeys, color=hf_c_dict)

    network.clear()
    network.draw_edges(color=uv_c_dict)

    # draw labels --------------------------------------------------------------
    if label:
        text_dict = {fkey: str(fkey) for fkey in hfkeys}
        volmesh.draw_facelabels(text=text_dict, color=hf_c_dict)
        network.draw_edgelabels(color=uv_c_dict)


def draw_directed_uv(datastructure,
                     uv_color=None):

    edges = []
    for u, v in datastructure.edges():

        sp = datastructure.vertex_coordinates(u)
        ep = datastructure.vertex_coordinates(v)

        edges.append({
            'start': sp,
            'end': ep,
            'arrow': 'end',
            'color': uv_color[(u, v)],
            'name': '{}.edge.{}-{}'.format(datastructure.name, u, v)})

    # 3. draw network ----------------------------------------------------------
    guids = compas_rhino.draw_lines(edges,
                                    layer=datastructure.layer,
                                    clear=False,
                                    redraw=False)

    return guids


def draw_directed_hf_and_uv(volmesh,
                            datastructure,
                            uv_color=None,
                            scale=1.0):

    hfkeys = pair_uv_to_hf(datastructure, volmesh).values()

    # 1. get colors ------------------------------------------------------------
    uv_colordict = color_to_colordict(uv_color,
                                      [uv for uv in datastructure.edges()],
                                      default=(0, 0, 0),
                                      colorformat='rgb',
                                      normalize=False)
    hf_colordict = get_force_colors_hf(volmesh,
                                       datastructure,
                                       uv_c_dict=uv_colordict)

    # 2. edges to draw ---------------------------------------------------------
    draw_directed_uv(datastructure, uv_color=uv_colordict)

    # 4. draw volmesh ----------------------------------------------------------
    volmesh.draw_faces(faces=hfkeys, color=hf_colordict)
    draw_volmesh_face_normals(volmesh,
                              hfkeys,
                              color=hf_colordict,
                              scale=scale)


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   other
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def draw_egi_arcs(egi):
    origin = egi.attributes['origin']

    for u, v in egi.edges():
        normal_1 = subtract_vectors(egi.vertex_coordinates(u), origin)
        normal_2 = subtract_vectors(egi.vertex_coordinates(v), origin)
        arc = draw_arc(normal_1, normal_2, origin)
        guid = add_curve(arc)
        obj = find_object(guid)
        color = (0, 0, 0)
        if egi.vertex[u]['type'] == 'zero' or egi.vertex[v]['type'] == 'zero':
            color = (255, 0, 0)
        attr = obj.Attributes
        attr.Name = egi.edge_name(u, v)
        attr.ObjectColor = FromArgb(*color)
        attr.ColorSource = ColorFromObject
        attr.ObjectDecoration = EndArrowhead
        obj.CommitChanges()


def bake_cells_as_polysurfaces(volmesh):
    for ckey in volmesh.cells():
        origin = volmesh.cell_centroid(ckey)
        ball = Sphere(Point3d(*tuple(origin)), 100)
        brep = Brep.CreateFromSphere(ball)
        rs.EnableRedraw(False)
        for hfkey in volmesh.cell_faces(ckey):
            center = volmesh.halfface_center(hfkey)
            normal = volmesh.halfface_normal(hfkey)
            plane = Plane(Point3d(*tuple(center)), Vector3d(*tuple(normal)))
            intersection = brep.Trim(plane, 0.1)
            if intersection:
                brep = brep.Trim(plane, 0.0001)[0]
                brep = brep.CapPlanarHoles(0.0001)
        guid = add_brep(brep)
        obj = find_object(guid)
        attr = obj.Attributes
        layer = 'cells_as_polysurfaces'
        rs.AddLayer(layer)
        index = find_layer_by_fullpath(layer, True)
        if index >= 0:
            attr.LayerIndex = index
        attr.Name = 'cell.{0}'.format(ckey)
        obj.CommitChanges()


def draw_arc(normal_1, normal_2, origin):
    mid_pt = normalize_vector(add_vectors(normal_1, normal_2))
    arc = Arc(Point3d(*[sum(axis) for axis in zip(normal_1, origin)]),
              Point3d(*[sum(axis) for axis in zip(mid_pt, origin)]),
              Point3d(*[sum(axis) for axis in zip(normal_2, origin)]))
    arc_as_curve = ArcCurve(arc)
    return arc_as_curve


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino

from compas_3gs.algorithms import volmesh_reciprocate

from compas_3gs.rhino import ReciprocationConduit


__commandname__ = "TGS_reciprocate"


def RunCommand(is_interactive):

    if '3GS' not in sc.sticky:
        compas_rhino.display_message('3GS has not been initialised yet.')
        return

    scene = sc.sticky['3GS']['scene']

    # get ForceVolMeshObject from scene
    objects = scene.find_by_name('force')
    if not objects:
        compas_rhino.display_message("There is no ForceDiagram in the scene.")
        return
    force = objects[0]

    # get ForceVolMeshObject from scene
    objects = scene.find_by_name('form')
    if not objects:
        compas_rhino.display_message("There is no FormDiagram in the scene.")
        return
    form = objects[0]

    # --------------------------------------------------------------------------

    current_setting = form.settings['show.nodes']
    if not current_setting:
        form.settings['show.nodes'] = True
        scene.update()

    # --------------------------------------------------------------------------

    # get global settings
    kmax = scene.settings['Solvers']['reciprocation.kmax']
    weight = scene.settings['Solvers']['reciprocation.alpha']
    refresh = scene.settings['Solvers']['reciprocation.refreshrate']

    options = ['Alpha', 'Iterations', 'Fix_vertices', 'Refreshrate']

    # --------------------------------------------------------------------------
    fix_vertices = []

    while True:
        option = compas_rhino.rs.GetString('Press Enter to run or ESC to exit.', strings=options)

        if option is None:
            print("Reciprocation aborted!")
            return

        if not option:
            break

        if option == 'Alpha':
            weight = compas_rhino.rs.GetReal(
                "Enter weight factor : 1  = form only... 0 = force only...", 1.0, minimum=0, maximum=1.0)

        if option == 'Iterations':
            new_kmax = compas_rhino.rs.GetInteger('Enter number of iterations', kmax, 1, 10000)
            if new_kmax or new_kmax is not None:
                kmax = new_kmax

        elif option == 'Fix_vertices':
            fix_vertices += form.select_vertices()

        elif option == 'RefreshRate':
            new_refresh = compas_rhino.rs.GetInteger('Refresh rate for dynamic visualisation', refresh, 0, 1000)
            if new_refresh or new_refresh is not None:
                refresh = new_refresh

    if refresh > kmax:
        refresh = 0

    scene.settings['Solvers']['reciprocation.kmax'] = kmax
    scene.settings['Solvers']['reciprocation.refreshrate'] = refresh

    # --------------------------------------------------------------------------

    if refresh > 0:

        conduit = ReciprocationConduit(force.diagram, form.diagram)

        def callback(forcediagram, formdiagram, k, args, refreshrate=refresh):
            if k % refreshrate:
                return
            conduit.redraw()

        with conduit.enabled():
            volmesh_reciprocate(force.diagram,
                                form.diagram,
                                weight=weight,
                                edge_min=scene.settings['Solvers']['reciprocation.l_min'],
                                edge_max=scene.settings['Solvers']['reciprocation.l_max'],
                                fix_vkeys=fix_vertices,
                                kmax=kmax,
                                tolerance=scene.settings['3GS']['tol.angles'],
                                callback=callback,
                                print_result_info=True)

    else:
        volmesh_reciprocate(force.diagram,
                            form.diagram,
                            kmax=kmax,
                            weight=weight,
                            edge_min=scene.settings['Solvers']['reciprocation.l_min'],
                            edge_max=scene.settings['Solvers']['reciprocation.l_max'],
                            fix_vkeys=fix_vertices,
                            tolerance=scene.settings['3GS']['tol.angles'],
                            callback=callback,
                            print_result_info=True)

    form.diagram.update_angle_deviations()

    form.settings['show.nodes'] = current_setting

    force.check_eq()
    form.check_eq()

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":

    RunCommand(True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division


__all__ = ['golden_section_search']


def golden_section_search(function, a, b, tol=1e-3):
    """Golden section search algorithm for finding the minimum or maximum of a strictly unimodal function
    by successively narrowing the range of values inside which the extremum is known to exist.

    Parameters
    ----------
    function : Function
        Evaluation function.
    a : float
        Lower bound of the initial search interval.
    b : float
        Upper bound of the initial search interval.
    tol : float, optional
        A tolerance for convergence.
        Default is ``1e-6``.

    Returns
    -------
    float
        The minimum or maximum value.

    Notes
    -----
    Adapted after [1]_.

    References
    ----------
    .. [1] Wikipedia *Golden section search*.
           Available at: https://en.wikipedia.org/wiki/Golden-section_search

    .. [2] Kiefer, J. (1953). Sequential minimax search for a maximum. In *Proceedings of the American Mathematical Society* 4(3).

    """

    gr = (5 ** 0.5 + 1) / 2

    c = b - (b - a) / gr
    d = a + (b - a) / gr

    while abs(c - d) > tol:

        if function(c) < function(d):
            b = d

        else:
            a = c

        c = b - (b - a) / gr
        d = a + (b - a) / gr

    return (b + a) / 2


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os

import System
import Eto.Drawing as drawing
import Eto.Forms as forms
import Rhino
import compas


class BrowserForm(forms.Form):

    def __init__(self, url=None, width=800, height=400):
        self.Owner = Rhino.UI.RhinoEtoApp.MainWindow
        self.Title = 'RhinoVault2'
        self.Padding = drawing.Padding(0)
        self.Resizable = False

        self.m_webview = forms.WebView()
        self.m_webview.Size = drawing.Size(width, height)

        if not url:
            url = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'src', 'index.html')
        self.m_webview.Url = System.Uri(url)
        self.m_webview.BrowserContextMenuEnabled = True

        layout = forms.DynamicLayout()
        layout.Spacing = drawing.Size(5, 5)
        layout.BeginVertical()
        layout.AddRow(self.m_webview)
        layout.EndVertical()
        self.Content = layout

        self.WindowStyle = forms.WindowStyle.None  # noqa E999
        self.m_webview.DocumentLoading += self.action
        if compas.WINDOWS:
            self.Location = drawing.Point(self.Owner.Location.X + self.Owner.Size.Width / 2 - 400, self.Owner.Location.Y + self.Owner.Size.Height / 2 - 200)

    def action(self, sender, e):
        if e.Uri.Scheme == "action" and e.Uri.Host == "close":
            self.Close()


def Browser():
    browserForm = BrowserForm()
    browserForm.Show()


if __name__ == "__main__":
    Browser()
